from __future__ import annotations

import logging
from bisect import bisect_left

from silx.gui import qt
from silx.io.url import DataUrl

from tomwer.core.scan.scanbase import TomwerScanBase

_logger = logging.getLogger(__name__)


class ManualFramesSelection(qt.QWidget):
    """Allows to select frame - angle to be used."""

    sigChanged = qt.Signal()
    """Signal emit when the frame selection changes"""

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self._anglesAvailable = []
        # cache of the angles available from the current defined frames. Must be sorted !!!
        self.setLayout(qt.QGridLayout())
        self.layout().addWidget(qt.QLabel("frame 1", self), 0, 0, 1, 1)
        self._frame1CB = qt.QComboBox(self)
        self._frame1CB.setEditable(True)
        self.layout().addWidget(self._frame1CB, 0, 1, 1, 1)

        self.layout().addWidget(qt.QLabel("frame 2", self), 1, 0, 1, 1)
        self._frame2CB = qt.QComboBox(self)
        self._frame2CB.setEditable(True)
        self.layout().addWidget(self._frame2CB, 1, 1, 1, 1)
        self._findAssociatedAnglePB = qt.QPushButton("+180°", self)
        button_180_font = self.font()
        button_180_font.setPixelSize(10)
        self._findAssociatedAnglePB.setFont(button_180_font)
        self._findAssociatedAnglePB.setFixedWidth(30)
        self._findAssociatedAnglePB.setSizePolicy(
            qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum
        )
        self.layout().addWidget(self._findAssociatedAnglePB, 1, 2, 1, 1)
        self._flipLRCB = qt.QCheckBox("flip L-R")
        self._flipLRCB.setChecked(True)
        self.layout().addWidget(self._flipLRCB, 1, 3, 1, 1)

        self._flipLRCB.toggled.connect(self._changed)
        self._frame1CB.currentIndexChanged.connect(self._changed)
        self._frame2CB.currentIndexChanged.connect(self._changed)
        self._findAssociatedAnglePB.released.connect(self._findAssociatedAngle)

    def _findAssociatedAngle(self):
        if self._frame1CB.count() == 0 or len(self._anglesAvailable) == 0:
            _logger.warning("no angles available, unable to get '+180°' frame")
        else:
            angle = float(self._frame1CB.currentText())
            # look for the closest 'associated' angle.
            # as the angles are not limited to [0-360] we need to check for any value.
            # if the angle is on the first part of the acquisition we expect to find it near angle +180
            # if it is on the second part (for 360 degree) we expect to find it on the first part (0-180)
            closest_pls_180_angle = self._getClosestAssociatedAngle(
                angle + 180.0, self._anglesAvailable
            )
            score_add = abs(closest_pls_180_angle - angle)
            closest_minus_180_angle = self._getClosestAssociatedAngle(
                angle - 180.0, self._anglesAvailable
            )
            score_sub = abs(closest_minus_180_angle - angle)
            if score_add >= score_sub:
                closest_180_angle = closest_pls_180_angle
            else:
                closest_180_angle = closest_minus_180_angle
            item_idx = self._frame2CB.findText(self._angleToStr(closest_180_angle))
            if item_idx < 0:
                _logger.error(f"Unable to find item for angle {closest_180_angle}")
            else:
                self._frame2CB.setCurrentIndex(item_idx)

    @staticmethod
    def _getClosestAssociatedAngle(angle: float, angles: tuple) -> float:
        """
        return the angle closest angle to 'angle' from 'angles'

        :warning: angles should be already sorted !!!
        """
        if angles is None or len(angles) == 0:
            return None
        if angle in angles:
            return angle
        pos = bisect_left(angles, angle)
        if pos == 0:
            return angles[0]
        elif pos > len(angles) - 1:
            return angles[-1]
        else:
            left_angle = angles[pos - 1]
            right_angle = angles[pos]
            if abs(right_angle - angle) > abs(left_angle - angle):
                return left_angle
            else:
                return right_angle

    def _changed(self):
        self.sigChanged.emit()

    @staticmethod
    def _angleToStr(angle: float) -> str:
        return f"{float(angle):0.3f}"

    def setScan(self, scan: TomwerScanBase | None) -> None:
        self._anglesAvailable.clear()
        self._frame1CB.clear()
        self._frame2CB.clear()
        if scan is None:
            return
        current_angle1 = self._getFrame1Angle()
        current_angle2 = self._getFrame2Angle()
        for proj_angle, proj_url in scan.get_proj_angle_url().items():
            try:
                angle = self._angleToStr(proj_angle)
            except Exception:
                angle = proj_angle
            else:
                self._anglesAvailable.append(float(proj_angle))
            self._frame1CB.addItem(angle, proj_url)
            self._frame2CB.addItem(angle, proj_url)

        self._anglesAvailable.sort()

        idx = self._frame1CB.findText(current_angle1)
        if idx >= 0:
            self._frame1CB.setCurrentIndex(idx)
        if current_angle1 == current_angle2:
            # if the two current angle are close then we consider it is better to look for angleX - angleX + 180 angles
            # instead of finding back angles
            self._findAssociatedAngle()
        else:
            idx = self._frame1CB.findText(current_angle1)
            if idx >= 0:
                self._frame1CB.setCurrentIndex(idx)

            idx = self._frame2CB.findText(current_angle2)
            if idx >= 0:
                self._frame2CB.setCurrentIndex(idx)

    def getFramesUrl(self, as_txt=False) -> tuple:
        """
        Return a tuple of (frame 1 url, frame 2 url). Url can be None
        """
        if as_txt:
            return self.getFrame1Url().path(), self.getFrame2Url().path()
        else:
            return self.getFrame1Url(), self.getFrame2Url()

    def getFrame1Url(self) -> DataUrl | None:
        return self._frame1CB.currentData()

    def getFrame2Url(self) -> DataUrl | None:
        return self._frame2CB.currentData()

    def _getFrame1Angle(self) -> str | None:
        return self._frame1CB.currentText()

    def _getFrame2Angle(self) -> str | None:
        return self._frame2CB.currentText()

    def isFrame2LRFLip(self):
        return self._flipLRCB.isChecked()



from orangewidget import gui, widget
from orangewidget.widget import Input, Output
from silx.gui import qt

from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.visualization.nxtomometadata import NXtomoMetadataViewer


class NXtomoMetadataViewerOW(widget.OWBaseWidget, openclass=True):
    """
    Widget to siplay NXtomo metadata
    """

    name = "nxtomo-metadata-viewer"
    id = "orange.widgets.tomwer.visualization.NXtomoMetadataViewerOW.NXtomoMetadataViewerOW"
    description = "Interface to display some metadata of a NXtomo"
    icon = "icons/nx_tomo_metadata_viewer.svg"
    priority = 13
    keywords = [
        "hdf5",
        "nexus",
        "tomwer",
        "file",
        "visualize",
        "visualization",
        "display",
        "NXTomo",
        "editor",
        "energy",
        "distance",
        "pixel size",
    ]

    class Inputs:
        data = Input(name="data", type=TomwerScanBase)

    class Outputs:
        data = Output(name="data", type=TomwerScanBase)

    want_main_area = True
    resizing_enabled = True

    def __init__(self, parent=None, *args, **kwargs):
        super().__init__(parent=parent, *args, **kwargs)

        _layout = gui.vBox(self.mainArea, self.name).layout()
        self.widget = NXtomoMetadataViewer(parent=self)
        _layout.addWidget(self.widget)

    def _validateScan(self, scan):
        self.Outputs.data.send(scan)
        super().hide()

    @Inputs.data
    def setScan(self, scan):
        if scan is None:
            pass
        elif not isinstance(scan, NXtomoScan):
            raise TypeError(
                f"expect to have an instance of {NXtomoScan}. {type(scan)} provided."
            )
        else:
            self.widget.setScan(scan)
            self.show()
            self.raise_()

    def sizeHint(self):
        return qt.QSize(400, 500)



from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.gui.control.series.nxtomoconcatenate import NXtomoConcatenateWidget


def test_NXtomoConcatenateWidget(
    qtapp,  # noqa F811
):
    """simple test of the NXtomoConcatenateWidget"""
    widget = NXtomoConcatenateWidget()
    assert widget.getConfiguration() == {
        "output_file": "{common_path}/concatenate.nx",
        "output_entry": "entry0000",
        "overwrite": False,
    }
    config = {
        "output_file": "helical.hdf5",
        "output_entry": "my_entry",
        "overwrite": True,
    }
    widget.setConfiguration(config=config)
    assert widget.getConfiguration() == config



# coding: utf-8
"""
material for radio and sinogram normalization
"""
from __future__ import annotations

from enum import Enum as _Enum
from tomoscan.normalization import Method


class _ValueSource(_Enum):
    MONITOR = "intensity monitor"
    MANUAL_ROI = "manual ROI"
    AUTO_ROI = "automatic ROI"
    DATASET = "from dataset"
    MANUAL_SCALAR = "scalar"
    NONE = "none"


class _ValueCalculationFct(_Enum):
    MEAN = "mean"
    MEDIAN = "median"


class _DatasetScope(_Enum):
    LOCAL = "local"
    GLOBAL = "global"


class _DatasetInfos:
    def __init__(self):
        self._scope = _DatasetScope.GLOBAL
        self._file_path = None
        self._data_path = None

    @property
    def scope(self) -> _DatasetScope:
        return self._scope

    @scope.setter
    def scope(self, scope: str | _DatasetScope):
        self._scope = _DatasetScope(scope)

    @property
    def file_path(self):
        return self._file_path

    @file_path.setter
    def file_path(self, file_path):
        self._file_path = file_path

    @property
    def data_path(self):
        return self._data_path

    @data_path.setter
    def data_path(self, data_path: str):
        self._data_path = data_path


class _ROIInfo:
    def __init__(self, x_min=None, x_max=None, y_min=None, y_max=None):
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max


class SinoNormalizationParams:
    """Information regarding the intensity normalization to be done"""

    def __init__(self, method=Method.NONE, source=_ValueSource.NONE, extra_infos=None):
        self._method = Method.NONE
        self._source = _ValueSource.NONE
        self._extra_infos = {}

        self.method = method
        self.extra_infos = extra_infos if extra_infos is not None else {}
        self.source = source

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, method: str | Method | None):
        if method is None:
            method = Method.NONE
        self._method = Method(method)

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source):
        if source is None:
            source = _ValueSource.NONE
        self._source = _ValueSource(source)

    @property
    def extra_infos(self):
        return self._extra_infos

    @extra_infos.setter
    def extra_infos(self, extra_infos: dict):
        if not isinstance(extra_infos, dict):
            raise TypeError("extra infos is expected to be a dictionary")
        elif "method" in extra_infos:
            raise KeyError("'method' is a reserved key name")
        else:
            self._extra_infos = extra_infos

    def to_dict(self):
        _dict = self._extra_infos
        _dict["method"] = self.method.value
        _dict["source"] = self.source.value
        return _dict

    @staticmethod
    def from_dict(dict_):
        params = SinoNormalizationParams()
        params.load_from_dict(dict_=dict_)
        return params

    def load_from_dict(self, dict_):
        tmp_dict = dict_.copy()
        if "method" in tmp_dict:
            self.method = tmp_dict["method"]
            del tmp_dict["method"]
        if "source" in tmp_dict:
            self.source = tmp_dict["source"]
            del tmp_dict["source"]
        self.extra_infos = tmp_dict



from __future__ import annotations

from silx.gui import qt

from ._ReconstructionParameters import ReconstructionParameters
from tomwer.gui.utils.illustrations import _IllustrationWidget


class GeometryOrMetadataWidget(qt.QTabWidget):
    """Simple widget that display either esrf geometric reference or volume metadata"""

    def __init__(self, parent=...):
        super().__init__(parent)

        # reconstruction parameters
        self._metadataWidget = ReconstructionParameters()
        self._metadataScrollArea = qt.QScrollArea(self)
        self._metadataScrollArea.setWidgetResizable(True)
        self._metadataScrollArea.setWidget(self._metadataWidget)

        self.addTab(self._metadataScrollArea, "volume metadata")

        # coordinate system illustration
        self._geometricRefLabel = _IllustrationWidget(self, img="3D_coordinate_system")
        self.addTab(self._geometricRefLabel, "geometry")

    # expos API
    def setVolumeIdentifier(self, *args, **kwargs):
        self._metadataWidget.setVolumeIdentifier(*args, **kwargs)

    def setMetadata(self, metadata: dict):
        self._metadataWidget.setVolumeMetadata(metadata=metadata)

    def sizeHint(self) -> qt.QSize:
        # size hint to match the plot2D one and have equal initial size
        return qt.QSize(500, 534)



from contextlib import contextmanager

from silx.gui import qt

if qt.BINDING == "PyQt5":
    from PyQt5.QtTest import QSignalSpy  # pylint: disable=E0401,E0611
elif qt.BINDING == "PySide6":
    from PySide6.QtTest import QSignalSpy  # pylint: disable=E0401,E0611
elif qt.BINDING == "PyQt6":
    from PyQt6.QtTest import QSignalSpy  # pylint: disable=E0401,E0611
else:
    QSignalSpy = None


@contextmanager
def block_signals(*objs):
    """Context manager blocking signals of QObjects.

    It restores previous state when leaving.

    :param qt.QObject objs: QObjects for which to block signals
    """
    blocked = [(obj, obj.blockSignals(True)) for obj in objs]
    try:
        yield
    finally:
        for obj, previous in blocked:
            obj.blockSignals(previous)



import os
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.process.output import ProcessDataOutputDirMode
from tomwer.core.process.control.nxtomomill import (
    EDFToNxProcess,
    H5ToNxProcess,
    get_default_raw_data_output_file,
)
from tomwer.core.utils.scanutils import MockNXtomo, MockEDF
from nxtomomill.converter.hdf5.utils import PROCESSED_DATA_DIR_NAME, RAW_DATA_DIR_NAME


def test_h52nx_process_deduce_output_file_path(tmp_path):
    """test H5ToNxProcess.deduce_output_file_path function"""
    scan_path = str(tmp_path / "path" / RAW_DATA_DIR_NAME / "my_scan")
    os.makedirs(scan_path)

    scan = MockNXtomo(scan_path=scan_path, n_proj=0).scan

    # test H52NXDefaultOutput.PROCESSED_DATA
    assert H5ToNxProcess.deduce_output_file_path(
        master_file_name=scan.master_file,
        scan=scan,
        output_dir=ProcessDataOutputDirMode.PROCESSED_DATA_FOLDER.value,
    ) == str(tmp_path / "path" / PROCESSED_DATA_DIR_NAME / "my_scan" / "my_scan.nx")

    # test H52NXDefaultOutput.NEAR_BLISS_FILE
    assert H5ToNxProcess.deduce_output_file_path(
        master_file_name=scan.master_file,
        scan=scan,
        output_dir=ProcessDataOutputDirMode.IN_SCAN_FOLDER.value,
    ) == str(tmp_path / "path" / RAW_DATA_DIR_NAME / "my_scan" / "my_scan.nx")

    # test providing output dir with some formatting to be done
    assert H5ToNxProcess.deduce_output_file_path(
        master_file_name=scan.master_file,
        scan=scan,
        output_dir="{scan_parent_dir_basename}/../../toto/{scan_dir_name}",
    ) == str(tmp_path / "toto" / "my_scan" / "my_scan.nx")

    # test providing output folder directly
    assert (
        H5ToNxProcess.deduce_output_file_path(
            master_file_name=scan.master_file,
            scan=scan,
            output_dir="/tmp/",
        )
        == "/tmp/my_scan.nx"
    )


def test_edf2nx_process_deduce_output_file_path(tmp_path):
    """test EDFToNxProcess.deduce_output_file_path function"""
    scan_path = str(tmp_path / "path" / RAW_DATA_DIR_NAME / "my_edf_scan")
    MockEDF(
        scan_path=scan_path,
        n_radio=10,
        n_ini_radio=10,
        n_extra_radio=0,
        dim=128,
        dark_n=1,
        flat_n=1,
    )
    scan = EDFTomoScan(scan_path)

    # test NEAR_INPUT_FILE
    assert EDFToNxProcess.deduce_output_file_path(
        folder_path=scan_path,
        output_dir=ProcessDataOutputDirMode.IN_SCAN_FOLDER.value,
        scan=scan,
    ) == os.path.join(tmp_path, "path", RAW_DATA_DIR_NAME, "my_edf_scan.nx")

    # test PROCESSED_DATA
    assert EDFToNxProcess.deduce_output_file_path(
        folder_path=scan_path,
        output_dir=ProcessDataOutputDirMode.PROCESSED_DATA_FOLDER.value,
        scan=scan,
    ) == os.path.join(tmp_path, "path", PROCESSED_DATA_DIR_NAME, "my_edf_scan.nx")

    # test providing output dir with some formatting to be done
    assert EDFToNxProcess.deduce_output_file_path(
        folder_path=scan_path,
        output_dir="{scan_parent_dir_basename}/../../toto/",
        scan=scan,
    ) == str(tmp_path / "toto" / "my_edf_scan.nx")

    # test providing output folder directly
    assert (
        EDFToNxProcess.deduce_output_file_path(
            folder_path=scan_path,
            output_dir="/tmp/output",
            scan=scan,
        )
        == "/tmp/output/my_edf_scan.nx"
    )


def test_get_default_raw_data_output_file():
    """test 'get_default_raw_data_output_file' function"""
    assert get_default_raw_data_output_file("/tmp/path/file.h5") == "/tmp/path/file.nx"
    assert (
        get_default_raw_data_output_file(f"/tmp/{PROCESSED_DATA_DIR_NAME}/file.h5")
        == f"/tmp/{RAW_DATA_DIR_NAME}/file.nx"
    )

    assert (
        get_default_raw_data_output_file(
            f"/tmp/path/{PROCESSED_DATA_DIR_NAME}/toto/file.h5"
        )
        == f"/tmp/path/{RAW_DATA_DIR_NAME}/toto/file.nx"
    )
    # note: _RAW_DATA_DIR_NAME part of the path but not a folder
    assert (
        get_default_raw_data_output_file(f"/tmp/path_{RAW_DATA_DIR_NAME}/toto/file.h5")
        == f"/tmp/path_{RAW_DATA_DIR_NAME}/toto/file.nx"
    )

    # 2. advance test
    # 2.1 use case: '_RAW_DATA_DIR_NAME' is present twice in the path -> replace the deeper one
    assert (
        get_default_raw_data_output_file(
            f"/tmp/{PROCESSED_DATA_DIR_NAME}/path/{RAW_DATA_DIR_NAME}/toto/file.h5"
        )
        == f"/tmp/{PROCESSED_DATA_DIR_NAME}/path/{RAW_DATA_DIR_NAME}/toto/file.nx"
    )

    # 2.2 use case: contains both '_RAW_DATA_DIR_NAME' and '_PROCESSED_DATA_DIR_NAME' in the path
    assert (
        get_default_raw_data_output_file(
            f"/tmp/{PROCESSED_DATA_DIR_NAME}/path/{RAW_DATA_DIR_NAME}/toto/file.h5"
        )
        == f"/tmp/{PROCESSED_DATA_DIR_NAME}/path/{RAW_DATA_DIR_NAME}/toto/file.nx"
    )

    assert (
        get_default_raw_data_output_file(
            f"/tmp/{RAW_DATA_DIR_NAME}/path/{PROCESSED_DATA_DIR_NAME}/toto/file.h5"
        )
        == f"/tmp/{RAW_DATA_DIR_NAME}/path/{RAW_DATA_DIR_NAME}/toto/file.nx"
    )

    # 2.3 use case: expected output file is the input file. Make sure append '_nxtomo'
    assert (
        get_default_raw_data_output_file("/tmp/path/file.nx")
        == "/tmp/path/file_nxtomo.nx"
    )



from __future__ import annotations

import os
import time
import logging

import numpy
import numpy.lib.npyio

from silx.gui import qt
from silx.gui.dialog.ColormapDialog import DisplayMode
from silx.gui.plot.ImageStack import ImageStack as _ImageStack
from silx.gui.plot.ImageStack import UrlLoader
from silx.io.url import DataUrl
from tomwer.core.utils.ftseriesutils import get_vol_file_shape
from tomwer.gui import icons
from tomwer.gui.visualization.fullscreenplot import FullScreenPlot2D
from tomwer.gui.visualization.reconstructionparameters import ReconstructionParameters
from tomwer.gui.visualization.scanoverview import ScanOverviewWidget
from tomwer.io.utils.utils import get_slice_data
from tomwer.core.scan.scanbase import TomwerScanBase

try:
    from PIL import Image
except ImportError:  # pragma: no cover
    has_PIL = False  # pragma: no cover
else:
    has_PIL = True

_logger = logging.getLogger(__name__)


class ImageStack(_ImageStack):
    """
    Image stack dedicated to data display.

    It deal for example with data normalization...
    """

    def __init__(self, parent, show_overview=True, backend=None):
        self._sliceMetadata: dict[str, dict] = {}
        # metadata associated to the slice. Key is the url (as str) and value is the dict containing metadata
        self._urlToScan: dict[str, TomwerScanBase] = {}
        # dict to link each slice to a scan id (in the case slices are raw projections - to set _scanOverviewDockerWidget)

        self._normFct = None
        self._url_indexes = None
        super().__init__(parent)
        self.getPlotWidget().setBackend(backend)
        self.getPlotWidget().setKeepDataAspectRatio(True)

        # tune colormap dialog to have histogram by default
        colormapAction = self.getPlotWidget().getColormapAction()
        colormapDialog = colormapAction.getColormapDialog()
        colormapDialog.getHistogramWidget().setDisplayMode(DisplayMode.HISTOGRAM)
        colormapAction.setColormapDialog(colormapDialog)

        self.setUrlLoaderClass(_TomwerUrlLoader)
        # hide axis to be display
        self._plot.setAxesDisplayed(False)
        self._loadSliceParams = False
        self._resetZoom = True

        # add dock widget for reconstruction parameters
        self._reconsInfoDockWidget = qt.QDockWidget(parent=self)
        self._reconsWidgetScrollArea = qt.QScrollArea(self)
        self._reconsWidgetScrollArea.setWidgetResizable(True)
        self._reconsWidget = ReconstructionParameters(self)
        self._reconsWidgetScrollArea.setWidget(self._reconsWidget)
        self._reconsInfoDockWidget.setWidget(self._reconsWidgetScrollArea)
        self._reconsInfoDockWidget.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._reconsInfoDockWidget)

        # add scan overview dock widget
        if show_overview:
            self._scanOverviewDockerWidget = qt.QDockWidget(parent=self)
            self._scanOverviewDockerWidget.setMaximumHeight(300)
            self._scanOverviewWidget = ScanOverviewWidget(self)
            self._scanOverviewDockerWidget.setWidget(self._scanOverviewWidget)
            self._scanOverviewDockerWidget.setFeatures(qt.QDockWidget.DockWidgetMovable)
            self.addDockWidget(
                qt.Qt.RightDockWidgetArea, self._scanOverviewDockerWidget
            )
        else:
            self._scanOverviewWidget = None
            self._scanOverviewDockerWidget = None

        # add search
        self._searchWidget = qt.QLineEdit(parent=self)
        self._searchWidget.setPlaceholderText("search")
        search_icon = icons.getQIcon("search")
        self._searchWidget.addAction(search_icon, qt.QLineEdit.LeadingPosition)
        # reorder urlsTable (include search QLineEdit)
        container = qt.QWidget()
        container_layout = qt.QGridLayout(container)

        container_layout.addWidget(self._searchWidget, 0, 1, 1, 2)
        container_layout.addWidget(self._urlsTable._toggleButton, 1, 2, 1, 1)
        container_layout.addWidget(self._urlsTable, 1, 1, 1, 2)
        self._urlsTable._urlsTable.model()
        self._tableDockWidget.setWidget(container)

        # for now hide the "control"
        self._urlsTable._toggleButton.hide()

        # add an action to plot the url-image 'full size'
        fullScreenIcon = icons.getQIcon("full_screen")
        self._fullScreenAction = qt.QAction(fullScreenIcon, "pop up full screen")
        self.getPlotWidget().toolBar().addAction(self._fullScreenAction)
        self._fullScreenAction.triggered.connect(self._popCurrentImageFullScreen)

        # set up
        self.setAutoResetZoom(False)

        # connect signal / slot
        self.sigCurrentUrlChanged.connect(self.updateAllMetadataDisplayed)
        self._searchWidget.textChanged.connect(self._filterUrlList)

    def _filterUrlList(self, *args, **kwargs):
        """Filter the url list according to the 'searchWidget'"""
        url_list = self._urlsTable._urlsTable
        filter_str = self._searchWidget.text()
        filter_str = filter_str.lstrip(" ").rstrip(" ")
        for item_index in range(url_list.count()):
            item = url_list.item(item_index)
            item.setHidden(filter_str not in item.text())

    def _popCurrentImageFullScreen(self, *args, **kwargs):
        new_plot = FullScreenPlot2D()
        active_image = self.getPlotWidget().getActiveImage()
        if active_image is None or active_image.getData(copy=False) is None:
            return
        url = self.getCurrentUrl()
        window_title = f"Plot of {url.path()}" if url is not None else "Image Plot"

        new_plot.setWindowTitle(window_title)
        # reuse the same colormap for convenience (user modification on it will be applied everywhere)
        new_plot.setDefaultColormap(self.getPlotWidget().getDefaultColormap())

        # add the current image
        new_plot.addImage(active_image.getData(copy=True))

        new_plot.showFullScreen()

    def setScan(self, scan):
        if self._scanOverviewWidget is not None:
            self._scanOverviewWidget.setScan(scan=scan)

    def getUrlListDockWidget(self):
        return self._tableDockWidget

    def resetZoom(self):
        self._plot.resetZoom()

    def setLimits(self, x_min, x_max, y_min, y_max):
        self._plot.setLimits(xmin=x_min, ymin=y_min, xmax=x_max, ymax=y_max)

    def getLimits(self):
        limits = []
        limits.extend(self._plot.getGraphXLimits())
        limits.extend(self._plot.getGraphYLimits())
        return tuple(limits)

    def setSliceReconsParamsVisible(self, visible):
        """show or not information regarding the slice reconstructed"""
        self._reconsInfoDockWidget.setVisible(visible)
        self._loadSliceParams = visible

    def setScanOverviewVisible(self, visible):
        """show or not overview of the scan"""
        if self._scanOverviewDockerWidget is not None:
            self._scanOverviewDockerWidget.setVisible(visible)

    def setNormalizationFct(self, fct, url_indexes=None):
        self._normFct = fct
        self._url_indexes = url_indexes

    def _urlLoaded(self) -> None:
        """

        :param url: result of DataUrl.path() function
        :return:
        """
        sender = self.sender()
        url = sender.url.path()
        if self._urlIndexes is not None and url in self._urlIndexes:
            data = sender.data
            if data is None:
                _logger.warning("no data found (is the url valid ?) " + url)
                return

            if data.ndim != 2:
                if data.ndim == 3:
                    if data.shape[0] == 1:
                        # if reconstruction along z
                        data = data.reshape((data.shape[1], data.shape[2]))
                    elif data.shape[1] == 1:
                        # if reconstruction along y
                        data = data.reshape((data.shape[0], data.shape[2]))
                    elif data.shape[2] == 1:
                        # if reconstruction along z
                        data = data.reshape((data.shape[0], data.shape[1]))
                    else:
                        _logger.warning(f"Image Stack only manage 2D data. Url: {url}")
                        return
                else:
                    _logger.warning(f"Image Stack only manage 2D data. Url: {url}")
                    return
            if self._normFct is None:
                self._urlData[url] = data
            else:
                norm_data = self._normFct(data, index=self._urlIndexes[url])
                self._urlData[url] = norm_data

            if self.getCurrentUrl().path() == url:
                self._plot.addImage(self._urlData[url])
                if hasattr(self, "getWaiterOverlay"):
                    self.getWaiterOverlay().hide()
                else:
                    self._waitingOverlay.hide()
                if self._resetZoom:
                    self._resetZoom = False
                    self._plot.resetZoom()

            if sender in self._loadingThreads:
                self._loadingThreads.remove(sender)
            self.sigLoaded.emit(url)

    def setResetZoomOnNextIteration(self, reset):
        self._resetZoom = reset

    def setCurrentUrl(self, url: DataUrl):
        if url in ("", None):
            url = None
        elif isinstance(url, str):
            url = DataUrl(path=url)
        elif not isinstance(url, DataUrl):
            raise TypeError
        if url is None:
            pass
        elif self._loadSliceParams:
            try:
                self.setCurrentSliceMetadata(self._metadata.get(url.path(), None))
            except Exception:
                _logger.info(f"Failed to find any metadata for {url.path()}.")
        super().setCurrentUrl(url)

    def setUrls(self, urls: list):
        self.clearSliceMetadata()
        _ImageStack.setUrls(self, urls)
        listWidget = self._urlsTable._urlsTable
        items = []
        for i in range(listWidget.count()):
            # TODO: do this on the fly
            item = listWidget.item(i)
            try:
                url = DataUrl(path=item.text())
            except Exception:
                _logger.info(
                    f"fail to deduce data of last modification for {item.text()}"
                )
            else:
                if os.path.exists(url.file_path()):
                    lst_m = time.ctime(os.path.getmtime(url.file_path()))
                    item.setToolTip(f"last modification : {lst_m}")
            items.append(listWidget.item(i))

    def updateAllMetadataDisplayed(self):
        # update widget related to image metadata.
        current_url = self.getCurrentUrl()
        if current_url is not None:
            current_url = current_url.path()

        self.setCurrentSliceMetadata(self._sliceMetadata.get(current_url, None))
        current_scan = self._urlToScan.get(current_url, None)
        if self._scanOverviewWidget:
            self._scanOverviewWidget.setScan(current_scan)

    def clearSliceMetadata(self):
        self._sliceMetadata.clear()

    def setCurrentSliceMetadata(self, metadata: dict | None) -> None:
        self._reconsWidget.setVolumeMetadata(metadata)

    def setSliceMetadata(self, metadata: dict[str, dict]) -> None:
        self._sliceMetadata = metadata
        self.updateAllMetadataDisplayed()

    def updateSliceMetadata(self, metadata: dict[str, dict]) -> None:
        self._sliceMetadata.update(metadata)
        self.updateAllMetadataDisplayed()

    def updateScanToUrl(self, scan_to_url: dict):
        self._urlToScan.update(scan_to_url)
        self.updateAllMetadataDisplayed()

    # scan metadata loading
    def reset(self):
        super().reset()
        self.clearSliceMetadata()
        self._urlToScan.clear()

    def _freeLoadingThreads(self):
        # overwrite it because the default waiting time seems very low
        # and make the CI fails
        for thread in self._loadingThreads:
            thread.blockSignals(True)
            thread.wait(100)
        self._loadingThreads.clear()


class _TomwerUrlLoader(UrlLoader):
    """
    Thread use to load DataUrl
    """

    def run(self):
        if self.url.file_path().endswith(".vol"):
            self.data = self._load_vol()
        elif self.url.scheme() == "tomwer":
            if has_PIL:
                self.data = numpy.array(Image.open(self.url.file_path()))
                if self.url.data_slice() is not None:
                    self.data = self.data[self.url.data_slice()]
            else:
                _logger.warning(
                    "need to install Pillow to read file " + self.url.file_path()
                )
                self.data = None
        else:
            try:
                self.data = get_slice_data(self.url)
            except IOError:
                self.data = None
            except ValueError:
                self.data = None
                _logger.warning(
                    f"Fail to open {self.url.path()}. Maybe the reconstruction failed."
                )
            except Exception as e:
                self.data = None
                _logger.error(
                    "Fail to load data from %s. Error is %s",
                    self.url.path(),
                    e,
                )

    def _load_vol(self):
        """
        load a .vol file
        """
        if self.url.file_path().lower().endswith(".vol.info"):
            info_file = self.url.file_path()
            raw_file = self.url.file_path().replace(".vol.info", ".vol")
        else:
            assert self.url.file_path().lower().endswith(".vol")
            raw_file = self.url.file_path()
            info_file = self.url.file_path().replace(".vol", ".vol.info")

        if not os.path.exists(raw_file):
            data = None
            mess = f"Can't find raw data file {raw_file} associated with {info_file}"
            _logger.warning(mess)
        elif not os.path.exists(info_file):
            mess = f"Can't find info file {info_file} associated with {raw_file}"
            _logger.warning(mess)
            data = None
        else:
            shape = get_vol_file_shape(info_file)
            if None in shape:
                _logger.warning(f"Fail to retrieve data shape for {info_file}.")
                data = None
            else:
                try:
                    numpy.zeros(shape)
                except MemoryError:
                    data = None
                    _logger.warning(
                        f"Raw file %s is too large for being read {raw_file}"
                    )
                else:
                    data = numpy.fromfile(
                        raw_file, dtype=numpy.float32, count=-1, sep=""
                    )
                    try:
                        data = data.reshape(shape)
                    except ValueError:
                        _logger.warning(
                            f"unable to fix shape for raw file {raw_file}. Look for information in {info_file}"
                        )
                        try:
                            sqr = int(numpy.sqrt(len(data)))
                            shape = (1, sqr, sqr)
                            data = data.reshape(shape)
                        except ValueError:
                            _logger.info(
                                f"deduction of shape size for {raw_file} failed"
                            )
                            data = None
                        else:
                            _logger.warning(
                                f"try deducing shape size for {raw_file} might be an incorrect interpretation"
                            )
        if self.url.data_slice() is None:
            return data
        else:
            return data[self.url.data_slice()]



# coding: utf-8
from __future__ import annotations

from silx.gui import qt

from tomwer.gui import icons as tomwericons


class HistoryAction(qt.QAction):
    """
    Action displaying the history of finished scans
    """

    def __init__(self, parent):
        icon = tomwericons.getQIcon("history")
        qt.QAction.__init__(self, icon, "history", parent)
        self.setCheckable(True)


class ConfigurationAction(qt.QAction):
    """
    Action to show the configuration dialog
    """

    def __init__(self, parent):
        icon = tomwericons.getQIcon("parameters")
        qt.QAction.__init__(self, icon, "configuration", parent)
        self.setCheckable(True)


class ObservationAction(qt.QAction):
    """
    Action to show the observation dialog
    """

    def __init__(self, parent):
        icon = tomwericons.getQIcon("loop")
        qt.QAction.__init__(self, icon, "observations", parent)
        self.setCheckable(True)


class ControlAction(qt.QAction):
    """
    Action to control the datawatcher (see status and select folder)
    """

    def __init__(self, parent):
        icon = tomwericons.getQIcon("health")
        qt.QAction.__init__(self, icon, "control", parent)
        self.setCheckable(True)



# coding: utf-8
from __future__ import annotations

import logging

from orangewidget import gui, widget
from orangewidget.widget import Input

import tomwer.core.process.visualization.sinogramviewer
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.visualization.sinogramviewer import SinogramViewer

from ..utils import WidgetLongProcessing

logger = logging.getLogger(__name__)


class SinogramViewerOW(WidgetLongProcessing, widget.OWBaseWidget, openclass=True):
    """
    This widget can be used to compute and display a specific sinogram from an
    acquisition.
    """

    name = "sinogram viewer"
    id = "orange.widgets.tomwer.visualization.sinogramviewer"
    description = (
        "This widget can be used to compute and display a "
        "specific sinogram from an acquisition."
    )

    icon = "icons/sinogramviewer.png"
    priority = 5
    keywords = ["tomography", "sinogram", "radio"]

    ewokstaskclass = (
        tomwer.core.process.visualization.sinogramviewer._SinogramViewerPlaceHolder
    )

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    class Inputs:
        data = Input(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        widget.OWBaseWidget.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        self._box = gui.vBox(self.mainArea, self.name)
        self._viewer = SinogramViewer(parent=self)
        self._box.layout().addWidget(self._viewer)

        # connect signal / slot
        self._viewer.sigSinoLoadStarted.connect(self._startProcessing)
        self._viewer.sigSinoLoadEnded.connect(self._endProcessing)

    @Inputs.data
    def addLeafScan(self, scanID):
        if scanID is None:
            return
        self._viewer.setScan(scanID)

    def close(self):
        self._viewer.close()
        super().close()



"""contains volumes classes and function. Abstraction layer for scans, based on tomoscan volumes.
See https://gitlab.esrf.fr/tomotools/tomoscan
"""

from .edfvolume import EDFVolume  # noqa F401
from .hdf5volume import HDF5Volume  # noqa F401
from .jp2kvolume import JP2KVolume  # noqa F401
from .rawvolume import RawVolume  # noqa F401
from .tiffvolume import MultiTIFFVolume, TIFFVolume  # noqa F401



# coding: utf-8


from tomwer.core.process.visualization.slicestack import _SliceStackPlaceHolder


def test_slice_stack():
    process = _SliceStackPlaceHolder(
        inputs={
            "data": None,
        }
    )
    process.run()



from __future__ import annotations

from silx.gui import qt
from silx.gui.qt import inspect as qt_inspect
from silx.gui.plot import PlotWidget


class OverlayMixIn:
    """
    MixIn class for overlay widget.

    For usage examples refer to :class:`WaitingOverlay`, :class:`LabelOverlay` and :class:`ButtonOverlay`

    .. warning:: Any class inheriting from this mixin must also inherit from a QWidget.
    """

    def __init__(
        self,
        parent,
    ):
        self._alignment: qt.Qt.AlignmentFlag = qt.Qt.AlignCenter
        self._alignmentOffsets: tuple[int, int] = (0, 0)
        self._registerParent(parent=parent)

    def getAlignment(self) -> qt.Qt.AlignmentFlag:
        return self._alignment

    def setAlignment(self, alignment: qt.Qt.AlignmentFlag):
        self._alignment = alignment
        self.update()

    def getAlignmentOffsets(self) -> tuple[int, int]:
        return self._alignmentOffsets

    def setAlignmentOffsets(self, offsets: tuple[int, int]):
        self._alignmentOffsets = offsets

    def _listenedWidget(self, parent: qt.QWidget) -> qt.QWidget:
        """Returns widget to register event filter to according to parent"""
        if isinstance(parent, PlotWidget):
            return parent.getWidgetHandle()
        return parent

    def _backendChanged(self):
        self._listenedWidget(self.parent()).installEventFilter(self)
        self._resizeLater()

    def _registerParent(self, parent: qt.QWidget | None):
        if parent is None:
            return
        self._listenedWidget(parent).installEventFilter(self)
        if isinstance(parent, PlotWidget):
            parent.sigBackendChanged.connect(self._backendChanged)
        self._resize()

    def _unregisterParent(self, parent: qt.QWidget | None):
        if parent is None:
            return
        if isinstance(parent, PlotWidget):
            parent.sigBackendChanged.disconnect(self._backendChanged)
        self._listenedWidget(parent).removeEventFilter(self)

    def setParent(self, parent: qt.QWidget):
        self._unregisterParent(self.parent())
        super().setParent(parent)
        self._registerParent(parent)

    def _getGeometry(self) -> qt.QRect | None:
        """Return the top left corner of the geometry to set up the geometry"""

        parent = self.parent()
        if parent is None:
            return None

        overlaySize = self.sizeHint()
        if isinstance(parent, PlotWidget):
            offset = parent.getWidgetHandle().mapTo(parent, qt.QPoint(0, 0))
            canvasLeft, canvasTop, canvasWidth, canvasHeight = (
                parent.getPlotBoundsInPixels()
            )
            canvasLeft += offset.x()
            canvasTop += offset.y()
        else:
            canvasWidth = parent.size().width()
            canvasHeight = parent.size().height()
            canvasLeft = 0
            canvasTop = 0

        # calculate left position
        if self._alignment & qt.Qt.AlignTop:
            top = canvasTop
        elif self._alignment & qt.Qt.AlignBottom:
            top = canvasTop + canvasHeight - overlaySize.height()
        else:
            top = canvasTop + (canvasHeight - overlaySize.height()) / 2

        # calculate top position
        if self._alignment & qt.Qt.AlignLeft:
            left = canvasLeft
        elif self._alignment & qt.Qt.AlignRight:
            left = canvasLeft + canvasWidth - overlaySize.width()
        else:
            left = canvasLeft + (canvasWidth - overlaySize.width()) / 2

        topLeft = qt.QPoint(
            int(left + self._alignmentOffsets[0]),
            int(top + self._alignmentOffsets[1]),
        )
        return qt.QRect(
            topLeft,
            overlaySize,
        )

    def _resize(self):
        if not qt_inspect.isValid(self):
            return  # For _resizeLater in case the widget has been deleted

        rect = self._getGeometry()
        if rect is None:
            return

        self.setGeometry(rect)
        self.raise_()

    def _resizeLater(self):
        qt.QTimer.singleShot(0, self._resize)

    def eventFilter(self, watched: qt.QWidget, event: qt.QEvent):
        if event.type() == qt.QEvent.Resize:
            self._resize()
            self._resizeLater()  # Defer resize for the receiver to have handled it
        return super().eventFilter(watched, event)



from __future__ import annotations

import logging
import numpy
from functools import partial

from silx.gui import qt
from silx.gui.plot.ColorBar import ColorBarWidget as _ColorBarWidget

from tomoscan.volumebase import SliceTuple
from tomwer.gui.utils.qt_utils import block_signals

from .VolumeViewerWidget import VolumeViewerWidget
from .toolbar import ToolBar
from silx.gui.colors import Colormap

_logger = logging.getLogger(__name__)


class ColorBarWidget(_ColorBarWidget):
    def _activeImageChanged(self, previous, legend):
        # we are force to overwrite this function to avoid resetting the colormap when
        # image changes. Somehow just disconnection from `_disconnectPlot` is not enough
        pass


class VolumeViewerWindow(qt.QMainWindow):
    """
    Window dedicated to the volume reconstruction summary.
    It add a toolbar common to the three plots.
    """

    sigLoadVolume = qt.Signal(bool)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._window = VolumeViewerWidget()
        self._slicesSet = False

        # main widget
        self.setCentralWidget(self._window)

        # colormap
        self._colormap = Colormap(name="gray", vmin=None, vmax=None)
        for plot in self._window.getPlots().values():
            plot.setDefaultColormap(self._colormap)
            plot.sigActiveImageChanged.connect(self._onSliceLoaded)

        # colorbar
        self._colorbarWidget = ColorBarWidget(
            plot=self._window.getMasterPlot(), parent=self._window
        )
        # Make colorbar background white
        self._colorbarWidget.setAutoFillBackground(True)
        self._colorbarWidget.layout().setSizeConstraint(qt.QLayout.SetNoConstraint)
        self._window._horizontalBottomSplitter.insertWidget(0, self._colorbarWidget)

        # window toolbar
        self._toolbar = ToolBar(
            master_plot=self._window.getMasterPlot(),
            parent=self._window.getMasterPlot(),
            colorbar=self._colorbarWidget,
            plots=self._window.getPlots().values(),
        )
        self.addToolBar(self._toolbar)

        # set up
        self._toolbar.showCoordinatesSystemLabel.setChecked(False)
        self._toolbar.showToolbar.setChecked(False)

        # connect signal / slot
        self._toolbar._loadVolumeWidget.toggled.connect(self.sigLoadVolume)
        self._toolbar.getExtendXYPlotAction().triggered.connect(
            partial(self.extendPlot, "XY")
        )
        self._toolbar.getExtendXZPlotAction().triggered.connect(
            partial(self.extendPlot, "XZ")
        )
        self._toolbar.getExtendYZPlotAction().triggered.connect(
            partial(self.extendPlot, "YZ")
        )
        self._toolbar.getResetDisplayAction().triggered.connect(
            partial(self.extendPlot, None)
        )

    def setSlicesAndMetadata(
        self,
        slices: dict[SliceTuple, numpy.ndarray],
        metadata: dict,
        volume_shape: tuple[int],
    ):
        self._slicesSet = len(slices) > 0
        self._colorbarWidget.setColormap(
            self._colormap, data=numpy.concatenate(tuple(slices.values()))
        )
        self._window.setVolumeShape(shape=volume_shape)
        self._window.setSlices(slices=slices)
        self._window.setVolumeMetadata(metadata=metadata, volume_shape=volume_shape)
        # required, especially because silx has known issues with it when keeping aspect ration
        self._window.resetZooms()

    def setLoading(self, loading: bool):
        """Function to interrupt loading - when finished or failed."""
        self._window.setLoading(loading)

    # expose API
    def getDefaultColormap(self) -> Colormap:
        return self._colormap

    def clear(self):
        self._slicesSet = False
        self._window.clear()

    def isLoadingVolume(self) -> bool:
        return self._toolbar._loadVolumeWidget.isChecked()

    def setLoadingVolume(self, loading_volume: bool) -> None:
        with block_signals(self):
            self._toolbar._loadVolumeWidget.setChecked(loading_volume)

    def setLoadingVolumeOptionVisible(self, visible) -> None:
        self._toolbar.getLoadVolumeAction().setVisible(visible)

    def initVolumePreview(self, *args, **kwargs):
        # reset colormap range to be able to update it when the first slice is displayed
        # (and before the full colormap is loaded in memory)
        self._colormap.setVRange(None, None)
        self._window.initVolumePreview(*args, **kwargs)

    def _onSliceLoaded(self, previous, legend):
        """
        When a new slice is loaded and when the colormap min/max have not been set already we want to update the range
        from the current slice values
        """
        if not self._slicesSet:
            data = self.sender().getImage(legend).getData()
            vMin, vMax = self._colormap._computeAutoscaleRange(data=data)
            self._colormap.setVRange(vMin, vMax)
            with block_signals(self._colormap):
                self._colormap.setVRange(None, None)

    def extendPlot(self, plot_name):
        # FIXME: we are force to hide / show the colorbar else
        # when calling the `setSizes` function this class
        # messup with the layout / resizing... couldn't find the origin of it.
        visible = self._colorbarWidget.isVisible()
        self._colorbarWidget.setVisible(False)
        self._window.extendPlot(plot_name)
        self._colorbarWidget.setVisible(visible)

    def setVolume(self, volume):
        self._window.setVolume(volume=volume)



# coding: utf-8
from __future__ import annotations


import os
import tempfile
from glob import glob

from ewoks import execute_graph
from ewokscore.graph import analysis, load_graph
from ewokscore.graph.validate import validate_graph

from tomwer.core.utils.scanutils import HDF5MockContext
from tomwer.core.process.reconstruction.output import (
    PROCESS_FOLDER_RECONSTRUCTED_VOLUMES,
)

try:
    from nabu.pipeline.fullfield.reconstruction import (  # noqa F401
        FullFieldReconstructor,
    )
except ImportError:
    try:
        from nabu.pipeline.fullfield.local_reconstruction import (  # noqa F401
            ChunkedReconstructor,
        )
    except ImportError:
        has_nabu = False
    else:
        has_nabu = True
else:
    has_nabu = True


def test_simple_workflow_nabu():
    """Test the workflow: darkref -> axis -> nabu slices -> nabu volume"""

    with HDF5MockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"), n_proj=100
    ) as scan:
        # insure no cfg yet
        assert len(glob(os.path.join(scan.path, "*.cfg"))) == 0

        graph = load_graph(
            source={
                "nodes": [
                    {
                        "id": "darkref",
                        "task_type": "class",
                        "task_identifier": "tomwer.core.process.reconstruction.darkref.darkrefs.DarkRefs",
                    },
                    {
                        "id": "axis",
                        "task_type": "class",
                        "task_identifier": "tomwer.core.process.reconstruction.axis.axis.AxisTask",
                        "default_inputs": [
                            {
                                "name": "axis_params",
                                "value": {
                                    "MODE": "manual",
                                    "POSITION_VALUE": 0.2,
                                },
                            },
                        ],
                    },
                    {
                        "id": "nabu slices",
                        "task_type": "class",
                        "task_identifier": "tomwer.core.process.reconstruction.nabu.nabuslices.NabuSlices",
                        "dry_run": True,  # as this is a mock dataset avoid to reconstruct it and only check for the .cfg file created
                        "default_inputs": [
                            {
                                "name": "nabu_params",
                                "value": {"tomwer_slices": 2},
                            }
                        ],
                    },
                    {
                        "id": "nabu volume",
                        "task_type": "class",
                        "task_identifier": "tomwer.core.process.reconstruction.nabu.nabuvolume.NabuVolume",
                        "dry_run": True,
                        "default_inputs": [
                            {
                                "name": "dry_run",
                                "value": True,
                            }
                        ],
                    },
                ],
                "links": [
                    {
                        "source": "darkref",
                        "target": "axis",
                        "map_all_data": True,
                    },
                    {
                        "source": "axis",
                        "target": "nabu slices",
                        "data_mapping": [  # same as all arguments but just here to test both
                            {
                                "source_output": "data",
                                "target_input": "data",
                            },
                        ],
                    },
                    {
                        "source": "nabu slices",
                        "target": "nabu volume",
                        "map_all_data": True,
                    },
                ],
            }
        )
        assert graph.is_cyclic is False, "graph is expected to be acyclic"
        assert analysis.start_nodes(graph=graph.graph) == {
            "darkref"
        }, "start node should be a single task `darkref`"
        validate_graph(graph.graph)
        result = execute_graph(
            graph,
            inputs=[
                {"id": "darkref", "name": "data", "value": scan},
            ],
        )
        assert analysis.end_nodes(graph=graph.graph) == {
            "nabu volume"
        }, "should only have one result nodes"
        assert "data" in result, f"cannot find `nabu volume` in {result}"
        assert os.path.exists(
            os.path.join(
                scan.path, PROCESS_FOLDER_RECONSTRUCTED_VOLUMES, "nabu_cfg_files"
            )
        ), "nabu has not been executed (even in dry mode)"



from __future__ import annotations

import logging

from nabu.stitching import config as stitching_config
from nabu.stitching.utils import ShiftAlgorithm as StitchShiftAlgorithm
from silx.gui import qt

from tomwer.gui.utils.qt_utils import block_signals
from tomwer.gui.configuration.level import ConfigurationLevel
from tomwer.gui.utils.scrollarea import QComboBoxIgnoreWheel as QComboBox
from tomwer.gui.utils.scrollarea import QSpinBoxIgnoreWheel as QSpinBox


_logger = logging.getLogger(__name__)


class StitcherAxisParams(qt.QWidget):
    """
    Widget to define settings of shift search along one axis
    """

    sigConfigChanged = qt.Signal()
    """signal emit when the configuration is updated"""

    DEFAULT_WINDOW_SEARCH_SIZE = 400
    # default window size for shift search

    def __init__(self, axis, parent=None) -> None:
        super().__init__(parent)
        if not isinstance(axis, int):
            raise TypeError("axis is expected to be an int")
        assert axis in (0, 1, 2)
        self._axis = axis
        self.setLayout(qt.QGridLayout())
        self._shiftSearchMethodCB = QComboBox(parent=self)
        # shift algorithm
        shift_search_methods = [
            StitchShiftAlgorithm.NABU_FFT,
            StitchShiftAlgorithm.SKIMAGE,
            StitchShiftAlgorithm.ITK_IMG_REG_V4,
            StitchShiftAlgorithm.NONE,
        ]
        for method in shift_search_methods:
            self._shiftSearchMethodCB.addItem(method.value)
        self._methodLabel = qt.QLabel("shift search method", self)
        self.layout().addWidget(self._methodLabel, 0, 0, 1, 1)
        self.layout().addWidget(self._shiftSearchMethodCB, 0, 1, 1, 1)
        # stitching options
        self._windowSizeLabel = qt.QLabel("window size", self)
        self.layout().addWidget(self._windowSizeLabel, 1, 0, 1, 1)
        self._windowSizeSB = QSpinBox(self)
        self._windowSizeSB.setMinimum(1)
        self._windowSizeSB.setMaximum(9999999)
        self._windowSizeSB.setValue(self.DEFAULT_WINDOW_SEARCH_SIZE)
        self._windowSizeSB.setSuffix("px")
        self.layout().addWidget(self._windowSizeSB, 1, 1, 1, 1)
        tooltip = "size of the window to try to refine shift"
        self._windowSizeLabel.setToolTip(tooltip)
        self._windowSizeSB.setToolTip(tooltip)

        # filter options
        self._filteringGroup = _FilteringGroupBox("filter for shift search", self)
        self._filteringGroup.setCheckable(True)
        self._filteringGroup.setChecked(False)
        self._filteringGroup.setLowPassValue(1)
        self._filteringGroup.setHighPassValue(20)
        self.layout().addWidget(self._filteringGroup, 2, 0, 2, 3)

        # set up
        self._methodChanged()

        # connect signal / slot
        self._shiftSearchMethodCB.currentIndexChanged.connect(self._methodChanged)
        self._windowSizeSB.valueChanged.connect(self._notifyConfigChanged)
        self._filteringGroup.sigValueChanged.connect(self._notifyConfigChanged)

    @property
    def axis(self):
        return self._axis

    def getShiftSearchMethod(self) -> StitchShiftAlgorithm:
        return StitchShiftAlgorithm(self._shiftSearchMethodCB.currentText())

    def setShiftSearchMethod(self, method: StitchShiftAlgorithm | str) -> None:
        method = StitchShiftAlgorithm(method)
        idx = self._shiftSearchMethodCB.findText(method.value)
        self._shiftSearchMethodCB.setCurrentIndex(idx)

    def getConfiguration(self) -> dict:
        method = self.getShiftSearchMethod()
        param_opts = [
            f"{stitching_config.KEY_IMG_REG_METHOD}={method.value}",
            f"{stitching_config.KEY_WINDOW_SIZE}={self._windowSizeSB.value()}",
        ]
        if self._filteringGroup.isChecked():
            param_opts.append(
                f"{stitching_config.KEY_LOW_PASS_FILTER}='{self._filteringGroup.getLowPassValue()}'"
            )
            param_opts.append(
                f"{stitching_config.KEY_HIGH_PASS_FILTER}='{self._filteringGroup.getHighPassValue()}'"
            )

        param_as_str = ";".join(param_opts)
        return {
            stitching_config.STITCHING_SECTION: {
                f"axis_{self.axis}_params": param_as_str,
            },
        }

    def setConfiguration(self, config: dict) -> None:
        with block_signals(self):
            axis_params = config.get(stitching_config.STITCHING_SECTION, {}).get(
                f"axis_{self.axis}_params", ""
            )
            options = filter(
                lambda a: a not in (None, ""), axis_params.replace(",", ";").split(";")
            )
            for option in options:
                clean_option = option.rstrip(" ").lstrip(" ")
                opt_name, opt_value = clean_option.split("=")
                if opt_value == "":
                    continue
                if opt_name == stitching_config.KEY_IMG_REG_METHOD:
                    self.setShiftSearchMethod(opt_value)
                elif opt_name == stitching_config.KEY_WINDOW_SIZE:
                    self._windowSizeSB.setValue(int(opt_value))
                elif opt_name == stitching_config.KEY_LOW_PASS_FILTER:
                    self._filteringGroup.setLowPassValue(int(opt_value))
                    self._filteringGroup.setEnabled(True)
                elif opt_name == stitching_config.KEY_HIGH_PASS_FILTER:
                    self._filteringGroup.setHighPassValue(int(opt_value))
                    self._filteringGroup.setEnabled(True)
                elif opt_name == stitching_config.KEY_OVERLAP_SIZE:
                    pass
                else:
                    _logger.error(f"option {opt_name} is not recognized")
        self._notifyConfigChanged()

    def _methodChanged(self) -> None:
        method = self.getShiftSearchMethod()
        with block_signals(self):
            self._windowSizeSB.setVisible(method is not StitchShiftAlgorithm.NONE)
            self._windowSizeLabel.setVisible(method is not StitchShiftAlgorithm.NONE)
        self._notifyConfigChanged()

    def _notifyConfigChanged(self, *args, **kwargs) -> None:
        self.sigConfigChanged.emit()

    def setConfigurationLevel(self, level: ConfigurationLevel):
        method = self.getShiftSearchMethod()
        self._windowSizeLabel.setVisible(
            level >= ConfigurationLevel.OPTIONAL
            and method is not StitchShiftAlgorithm.NONE
        )
        self._windowSizeSB.setVisible(
            level >= ConfigurationLevel.OPTIONAL
            and method is not StitchShiftAlgorithm.NONE
        )
        self._filteringGroup.setVisible(level >= ConfigurationLevel.ADVANCED)


class _FilteringGroupBox(qt.QGroupBox):
    sigValueChanged = qt.Signal()
    """emit when one of the low or high pass value changed or when toggled"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setLayout(qt.QFormLayout())

        self._lowPassFilterSP = QSpinBox(self)
        self.layout().addRow("low pass filter", self._lowPassFilterSP)

        self._highPassFilterSP = QSpinBox(self)
        self.layout().addRow("high pass filter", self._highPassFilterSP)

        # connect signal / slot
        self.toggled.connect(self._updateEnability)
        self.toggled.connect(self._valueChanged)
        self._lowPassFilterSP.editingFinished.connect(self._valueChanged)
        self._highPassFilterSP.editingFinished.connect(self._valueChanged)

    def _updateEnability(self):
        self._lowPassFilterSP.setEnabled(self.isChecked())
        self._highPassFilterSP.setEnabled(self.isChecked())

    def getLowPassValue(self):
        return self._lowPassFilterSP.value()

    def setLowPassValue(self, value):
        self._lowPassFilterSP.setValue(value)

    def getHighPassValue(self):
        return self._highPassFilterSP.value()

    def setHighPassValue(self, value):
        self._highPassFilterSP.setValue(value)

    def _valueChanged(self, *args, **kwargs):
        self.sigValueChanged.emit()



from tomwer.gui.stitching.normalization import NormalizationBySampleGroupBox

from tomwer.tests.conftest import qtapp  # noqa F401


def test_FrameNormalizationWidget(
    qtapp,  # noqa F811
):
    widget = NormalizationBySampleGroupBox()
    assert widget.getConfiguration() == {
        "active": True,
        "side": "left",
        "method": "median",
        "width": 30,
        "margin": 0,
    }

    new_config = {
        "active": False,
        "side": "left",
        "method": "median",
        "width": 30,
        "margin": 0,
    }

    widget.setConfiguration(new_config)
    assert widget.getConfiguration() == new_config



from __future__ import annotations

import copy
import functools
import logging
import os
import gc
from collections.abc import Callable
from tomwer.io.utils import format_stderr_stdout
from silx.utils.deprecation import deprecated, deprecated_warning
from silx.io.utils import open as open_hdf5

from processview.core.manager.manager import ProcessManager, DatasetState

from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.utils.scanutils import data_identifier_to_scan
from tomwer.core.utils.slurm import is_slurm_available
from tomwer.core.process.reconstruction.nabu.utils import (
    update_nabu_config_for_tiff_3d,
    from_nabu_config_to_file_format,
)
from tomwer.core.process.reconstruction.nabu.plane import NabuPlane
from tomwer.core.process.reconstruction.nabu.utils import slice_index_to_int
from tomwer.core.process.reconstruction.utils.cor import relative_pos_to_absolute
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.core.process.reconstruction.output import (
    PROCESS_FOLDER_RECONSTRUCTED_SLICES,
)

try:
    from nabu.pipeline.fullfield.reconstruction import (  # noqa F401
        FullFieldReconstructor,
    )
except (ImportError, OSError) as e:
    _err_import_nabu = e
    try:
        from nabu.pipeline.fullfield.local_reconstruction import (  # noqa F401
            ChunkedReconstructor,
        )
    except (ImportError, OSError):
        # import of cufft library can bring an OSError if cuda not install
        has_nabu = False
    else:
        has_nabu = True
        _err_import_nabu = None
else:
    has_nabu = True
    _err_import_nabu = None

from typing import Iterable

from nabu import version as nabu_version
from nabu.pipeline.config import (
    _extract_nabuconfig_keyvals,
    generate_nabu_configfile,
    get_default_nabu_config,
)
from nabu.pipeline.fullfield.nabu_config import (
    nabu_config as nabu_fullfield_default_config,
)
from processview.core.superviseprocess import SuperviseProcess
from enum import Enum as _Enum

from tomwer.core.process.task import Task
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.utils import docstring

from . import settings as nabu_settings
from . import utils
from .nabucommon import (
    ResultsLocalRun,
    ResultSlurmRun,
    ResultsRun,
    _NabuBaseReconstructor,
)
from .target import Target

_logger = logging.getLogger(__name__)
if not has_nabu:
    _logger.error(_err_import_nabu)


def run_slices_reconstruction(
    scan: TomwerScanBase,
    config: dict,
    dry_run: bool = False,
    advancement=None,
    process_id: int | None = None,
    instantiate_classes_only: bool = False,
    invalid_slice_callback: Callable[[str], None] | None = None,
) -> tuple:
    """
    call nabu for a reconstruction on scan with the given configuration

    :param scan: scan to reconstruct
    :param config: configuration to run the reconstruction.
                        Contains nabu reconstruction parameters and slurm cluster
                        configuration if requested (key: `slurm-cluster`).
    :param dry_run: do we want to run dry
    :param local: do we want to run a local reconstruction
    :param stderr: file to redirect stderr
    :param stdout: file to redirect stdout
    :param advancement: optional Progress class to display advancement
    :param process_id: optional process id
    :param instantiate_class_only: if we don't want to run the SingleSliceRunner but only return them. Use case: we want to keep a hand on processing and it can be cancelled

    :return: (all_succeed, stdouts, stderrs, final_configs, future_scan)
        * all_succeed: bool, True if all the reconstruction succeed or if all job request succeed.
        * stdouts: list of stdout of job reconstruction or job requests
        * stderrs: list of stderr of job reconstruction or job requests
        * final_configs: list of configurations submits to nabu
        * future_scan: FutureTomwerScan | None future scan containing futures pointing to job submited to the cluster. None if local reconstruction

    Behavior: will clear the last slices reconstructed
    """
    # TODO: remove the local parameter
    _logger.info(f"start reconstruction of {scan}")

    cluster_config = config.pop("cluster_config", None)
    if cluster_config == {}:
        cluster_config = None
    is_cluster_job = cluster_config is not None
    if is_cluster_job and not is_slurm_available():
        raise ValueError(
            "job on cluster requested but no access to slurm cluster found"
        )

    # beam shape is not directly used by nabu (uses ctf_geometry directly)
    config.get("phase", {}).pop("beam_shape", None)

    # if scan contains some center of position copy it to nabu
    if scan.axis_params is not None and scan.axis_params.relative_cor_value is not None:
        if "reconstruction" in config:
            # move the cor value to the nabu reference
            cor_nabu_ref = relative_pos_to_absolute(
                relative_pos=scan.axis_params.relative_cor_value,
                det_width=scan.dim_1,
            )
            config["reconstruction"]["rotation_axis_position"] = str(cor_nabu_ref)
    _logger.info(f"set nabu reconstruction parameters to {scan}")

    # update nabu recons_params used
    sc_config = get_default_nabu_config(nabu_fullfield_default_config)
    sc_config.update(config)
    scan.nabu_recons_params = sc_config

    # handle special cases like several db...
    nabu_configurations = interpret_tomwer_configuration(
        config, scan=scan, invalid_slice_callback=invalid_slice_callback
    )
    output_urls = []
    stderrs = []
    stdouts = []
    final_configs = []
    futures = []
    instantiated_classes = []
    all_succeed = True
    if advancement is not None:
        advancement.total = len(nabu_configurations)
    scan.clear_latest_reconstructions()
    for nabu_configuration in nabu_configurations:
        l_config, slice_index = nabu_configuration
        result = run_single_slice_reconstruction(
            nabu_config=l_config,
            cluster_config=cluster_config,
            scan=scan,
            slice_index=slice_index,
            dry_run=dry_run,
            instantiate_class_only=instantiate_classes_only,
            axis=config.get("reconstruction", {}).get("slice_plane", "XY"),
        )

        # specific treatments of results
        if result is None:
            # in case of timeout or another issue. Log should already have been provided
            pass
        elif instantiate_classes_only:
            instantiated_classes.append(result)
            continue
        if slice_index is None:
            continue
        elif isinstance(result, ResultsLocalRun):
            assert not is_cluster_job, "cluster job should not return ResultsLocalRun"
            stderrs.append(result.std_err)
            stdouts.append(result.std_out)
            output_urls.extend(result.results_identifiers)
            # if slice_index is None this mean that we are simply creating the
            # .cfg file for nabu full volume.
        elif isinstance(result, ResultSlurmRun):
            assert (
                is_cluster_job
            ), "local reconstruction should not return ResultSlurmRun"
            stderrs.append(result.std_err)
            stdouts.append(result.std_out)
            futures.extend(result.future_slurm_jobs)
        elif not isinstance(result, ResultsRun):
            raise ValueError(
                f"result is expected to be an instance of {ResultsRun} not {type(result)}"
            )

        # common treatments of results
        if result is not None:
            final_configs.append(result.config)
            all_succeed = all_succeed and result.success

        if advancement is not None:
            advancement.update()

    if instantiate_classes_only:
        return instantiated_classes
    if is_cluster_job:
        future_tomo_obj = FutureTomwerObject(
            tomo_obj=scan,
            futures=tuple(futures),
            process_requester_id=process_id,
        )
        scan.set_latest_reconstructions(output_urls)
        return all_succeed, stdouts, stderrs, final_configs, future_tomo_obj
    else:
        # tag latest reconstructions
        scan.set_latest_reconstructions(output_urls)
        return all_succeed, stdouts, stderrs, final_configs, None


class NabuSlicesTask(
    Task,
    SuperviseProcess,
    input_names=(
        "data",
        "nabu_params",
    ),
    optional_input_names=(
        "dry_run",
        "serialize_output_data",
        "invalid_slice_callback",
    ),
    output_names=("data", "nabu_params", "future_tomo_obj"),
):
    """
    Definition of the nabu reconstruction volume reconstruction process
    """

    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        SuperviseProcess.__init__(self, process_id=process_id)
        Task.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        self._dry_run = inputs.get("dry_run", False)
        self._current_processing = None
        # we can sometime call several time a nabu subprocess. The idea is to keep track of it
        # if we want to stop the processing

    def run(self):
        scan = data_identifier_to_scan(self.inputs.data)
        configuration = self.inputs.nabu_params

        self.outputs.nabu_params = None
        if scan is None:
            self.outputs.data = None
            return
        if isinstance(scan, TomwerScanBase):
            scan = scan
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise ValueError(f"input type of {scan}: {type(scan)} is not managed" "")
        assert isinstance(configuration, dict), "configuration is expected to be a dict"

        output_urls = []
        stderrs = []
        stdouts = []
        final_configs = []
        futures = []
        all_succeed = True
        is_cluster_job = configuration.get("cluster_config", None) is not None
        # loop is required for distributed since version 2021
        try:
            single_slice_runner_instances = run_slices_reconstruction(
                config=configuration,
                scan=scan,
                dry_run=self._dry_run,
                process_id=self.process_id,
                instantiate_classes_only=True,
                invalid_slice_callback=self.get_input_value(
                    "invalid_slice_callback", None
                ),
            )
        except Exception as e:
            mess = f"Fail to instantiate slice reconstructor for {str(scan)}. Reason is {e}."
            _logger.processFailed(mess)
            ProcessManager().notify_dataset_state(
                dataset=scan,
                process=self,
                state=DatasetState.FAILED,
                details=mess,
            )
            self.outputs.future_tomo_obj = None
            return

        for slice_runner in single_slice_runner_instances:
            self._current_processing = slice_runner
            result = self._current_processing.run()[
                0
            ]  # we are expecting a single slice per run in this configuration
            if result is None:
                # in case of timeout or another issue. Log should already have been provided
                pass
            elif isinstance(result, ResultsLocalRun):
                assert (
                    not is_cluster_job
                ), "cluster job should not return ResultsLocalRun"
                stderrs.append(result.std_err)
                stdouts.append(result.std_out)
                output_urls.extend(result.results_identifiers)
                # if slice_index is None this mean that we are simply creating the
                # .cfg file for nabu full volume.
            elif isinstance(result, ResultSlurmRun):
                assert (
                    is_cluster_job
                ), "local reconstruction should not return ResultSlurmRun"
                stderrs.append(result.std_err)
                stdouts.append(result.std_out)
                futures.extend(result.future_slurm_jobs)
            elif not isinstance(result, ResultsRun):
                raise ValueError(
                    f"result is expected to be an instance of {ResultsRun} not {type(result)}"
                )

            # common treatments of results
            if result is not None:
                final_configs.append(result.config)
                all_succeed = all_succeed and result.success

        # update future object and scan latest reconstructions
        self._current_processing = None
        if not self._cancelled and is_cluster_job:
            future_tomo_obj = FutureTomwerObject(
                tomo_obj=scan,
                futures=tuple(futures),
                process_requester_id=self.process_id,
            )
            scan.set_latest_reconstructions(output_urls)
        else:
            # tag latest reconstructions
            scan.set_latest_reconstructions(output_urls)
            future_tomo_obj = None

        volume_urls = []
        for rec_identifier in scan.latest_reconstructions:
            volume_urls.extend(
                VolumeFactory.from_identifier_to_vol_urls(rec_identifier)
            )

        # update processes information / registration
        gc.collect()

        # TODO: check output files with the tomoscan validator ?
        if self._cancelled:
            state = DatasetState.CANCELLED
            details = "cancelled by user"
            future_tomo_obj = None
            _logger.info(f"Slices computation for {scan} cancelled")
        else:
            if not all_succeed:
                mess = f"Slices computation for {scan} failed."
                state = DatasetState.FAILED
                _logger.processFailed(mess)
            else:
                state = DatasetState.SUCCEED
                mess = f"Slices computed for {scan}."
                _logger.processSucceed(mess)
            elmts = [
                format_stderr_stdout(stderr=stderr, stdout=stdout)
                for stderr, stdout in zip(stderrs, stdouts)
            ]
            elmts.insert(0, mess)
            details = "\n".join(elmts)

        ProcessManager().notify_dataset_state(
            dataset=scan,
            process=self,
            state=state,
            details=details,
        )

        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan
        self.outputs.nabu_params = scan.nabu_recons_params
        self.outputs.future_tomo_obj = future_tomo_obj

    def set_configuration(self, configuration: dict) -> None:
        Task.set_configuration(self, configuration=configuration)
        if "dry_run" in configuration:
            self.set_dry_run(bool(configuration["dry_run"]))

    @staticmethod
    def program_name():
        return "nabu-slices"

    @staticmethod
    def program_version():
        return nabu_version

    @deprecated(replacement="provide dry_run to the task inputs", since_version="1.2")
    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    @property
    @deprecated(replacement="Task.inputs.dry_run", since_version="1.2")
    def dry_run(self):
        return self._dry_run

    def cancel(self):
        """
        stop current processing
        """
        self._cancelled = True
        if self._current_processing is not None:
            self._current_processing.cancel()


def interpret_tomwer_configuration(
    config: dict,
    scan: TomwerScanBase | None,
    invalid_slice_callback: Callable[[str], None] | None = None,
) -> tuple:
    """
    tomwer can 'mock' the nabu reconstruction to request more feature.
    Typical use case is that we can ask for reconstruction of several
    slices and not only the volume

    :param config: tomwer configuration for nabu
    :param scan: requested if we want to get slices
    :return: tuple of tuples (nabu configuration, is slice)
    """
    if not isinstance(config, dict):
        raise TypeError("config is expected to be a dict")

    def get_nabu_config(config):
        nabu_config = copy.deepcopy(config)
        if "tomwer_slices" in nabu_config:
            del nabu_config["tomwer_slices"]
        return nabu_config

    if "tomwer_slices" in config and scan is not None:
        reconstruction_axis = NabuPlane.from_value(
            config.get("reconstruction", {}).get("slice_plane", "XY")
        )
        slices = list(
            NabuSliceMode.getSlices(
                config["tomwer_slices"],
                scan=scan,
                axis=reconstruction_axis,
            )
        )

        def filter_slice(slice_index: int, axis: NabuPlane):
            """remove slices that 'cannot' be reconstructed (out of bounds)"""
            if axis is NabuPlane.XY:
                index_max = scan.dim_2
            elif axis in (NabuPlane.XZ, NabuPlane.YZ):
                index_max = scan.dim_1
            else:
                raise ValueError

            if index_max is None:
                return True

            index_max = index_max - 1

            if slice_index > index_max:
                error_message = f"{scan.get_identifier().short_description()}\n\nCannot reconstruct slice {slice_index} along plane {axis.value}. Must be in [0-{index_max}]"
                _logger.error(error_message)
                if invalid_slice_callback is not None:
                    invalid_slice_callback(error_message)
                return False
            return True

        slices = list(
            filter(
                lambda slice_index: filter_slice(
                    int(slice_index), axis=reconstruction_axis
                ),
                slices,
            )
        )

    else:
        slices = []

    if "phase" in config and "delta_beta" in config["phase"]:
        pag_dbs = config["phase"]["delta_beta"]
        if isinstance(pag_dbs, str):
            pag_dbs = utils.retrieve_lst_of_value_from_str(pag_dbs, type_=float)
        if len(pag_dbs) == 0:
            pag_dbs = (None,)
    else:
        pag_dbs = (None,)

    # by default add the slice 'None' which is the slice for the volume
    slices.append(None)
    nabu_config = get_nabu_config(config=config)
    res = []
    for slice_ in slices:
        for pag_db in pag_dbs:
            local_config = copy.deepcopy(nabu_config)
            if pag_db is not None:
                local_config["phase"]["delta_beta"] = str(pag_db)
            res.append((local_config, slice_))
    return tuple(res)


class SingleSliceRunner(_NabuBaseReconstructor):
    def __init__(
        self,
        scan: TomwerScanBase,
        config: dict,
        dry_run: bool,
        slice_index: int | str | None,
        axis: NabuPlane,
        target: Target,
        cluster_config: dict | None,
        process_name: str,
        add_to_latest_reconstructions: bool = True,
    ) -> None:
        super().__init__(
            scan=scan,
            dry_run=dry_run,
            target=target,
            cluster_config=cluster_config,
            process_name=process_name,
            axis=axis,
        )
        self._slice_index = slice_index
        if not isinstance(config, dict):
            raise TypeError(f"config is expected to be a dictionary not {type(dict)}")
        self._config = config
        self._add_to_latest_reconstructions = add_to_latest_reconstructions

    @property
    def slice_index(self):
        return self._slice_index

    @property
    def config(self):
        return self._config

    @property
    def add_to_latest_reconstructions(self):
        return self._add_to_latest_reconstructions

    @property
    def processed_data_folder_name(self):
        """return the specific processed folder name associated to this type of reconstruction."""
        return PROCESS_FOLDER_RECONSTRUCTED_SLICES

    @docstring(_NabuBaseReconstructor)
    def only_create_config_file(self):
        return self.slice_index is None

    @docstring(_NabuBaseReconstructor)
    def run(self) -> Iterable:
        """
        If the target is local will wait for the reconstruction to be finish to return

        :raise: TIMEOUT_SLURM_JOB_SUBMISSION if not all workers spwan
        """
        if isinstance(self.slice_index, str):
            self._slice_index = slice_index_to_int(
                slice_index=self.slice_index,
                scan=self.scan,
                axis=self.axis,
            )
        elif (
            isinstance(self.slice_index, float)
            and int(self.slice_index) == self.slice_index
        ):
            self._slice_index = int(self.slice_index)
        elif not isinstance(self.slice_index, (int, type(None))):
            raise TypeError(
                f"slice index is expected to an int or 'middle' or None and not {type(self.slice_index)}"
            )
        config_complete = _extract_nabuconfig_keyvals(nabu_fullfield_default_config)
        config_complete["dataset"] = self.scan.get_nabu_dataset_info()
        for key in config_complete.keys():
            if key in self.config:
                config_complete[key].update(self.config[key])
        config = config_complete

        config["resources"] = config.get("resources", {})
        config["resources"]["method"] = "local"

        # force overwrite results
        if "output" not in config:
            config["output"] = {}

        config["output"].update({"overwrite_results": 1})

        config, cfg_folder = self._treateOutputSliceConfig(config)
        # the policy is to save nabu .cfg file at the same location as the
        # force overwrite results
        if self.slice_index is not None:
            if self.axis is NabuPlane.YZ:
                config["reconstruction"]["start_x"] = self.slice_index
                config["reconstruction"]["end_x"] = self.slice_index
            elif self.axis is NabuPlane.XZ:
                config["reconstruction"]["start_y"] = self.slice_index
                config["reconstruction"]["end_y"] = self.slice_index
            elif self.axis is NabuPlane.XY:
                config["reconstruction"]["start_z"] = self.slice_index
                config["reconstruction"]["end_z"] = self.slice_index
            else:
                raise ValueError(
                    f"self.axis has an invalid value: {self.axis} when expected to be in {[item.value for item in NabuPlane]}"
                )

        if self.slice_index is not None:
            os.makedirs(config["output"]["location"], exist_ok=True)

        name = (
            config["output"]["file_prefix"] + nabu_settings.NABU_CONFIG_FILE_EXTENSION
        )
        if not isinstance(self.scan, EDFTomoScan):
            name = "_".join((self.scan.entry.lstrip("/"), name))
        conf_file = os.path.join(cfg_folder, name)

        _logger.info(f"{self.scan}: create {conf_file}")
        file_format = config["output"]["file_format"]
        update_nabu_config_for_tiff_3d(config)
        # add some tomwer metadata and save the configuration
        # note: for now the section is ignored by nabu but shouldn't stay that way
        with utils.TomwerInfo(config) as config_to_dump:
            generate_nabu_configfile(
                conf_file,
                nabu_fullfield_default_config,
                config=config_to_dump,
                options_level="advanced",
            )

        return tuple(
            [
                self._process_config(
                    config_to_dump=config_to_dump,
                    config_file=conf_file,
                    info="nabu slice reconstruction",
                    file_format=file_format,
                    process_name=self.process_name,
                ),
            ]
        )

    @docstring(_NabuBaseReconstructor)
    def _get_futures_slurm_callback(self, config_to_dump):
        if self.add_to_latest_reconstructions:
            # add callback to set slices reconstructed urls
            class CallBack:
                # we cannot create a future directly because distributed enforce
                # the callback to have a function signature with only the future
                # as single parameter.
                def __init__(self, f_partial, scan) -> None:
                    self.f_partial = f_partial
                    self.scan = scan

                def process(self, fn):
                    if fn.done() and not (fn.cancelled() or fn.exception()):
                        # update reconstruction urls only if processing succeed.
                        recons_urls = self.f_partial()
                        self.scan.add_latest_reconstructions(recons_urls)

            file_format = from_nabu_config_to_file_format(config_to_dump)
            callback = functools.partial(
                utils.get_recons_volume_identifier,
                file_prefix=config_to_dump["output"]["file_prefix"],
                location=config_to_dump["output"]["location"],
                file_format=file_format,
                scan=self.scan,
                slice_index=None,
                axis=self.axis,
            )

            return (CallBack(callback, self.scan),)
        else:
            return super()._get_futures_slurm_callback(config_to_dump)

    @staticmethod
    def get_file_basename_reconstruction(
        scan,
        pag,
        ctf,
        db,
        slice_index: str | int,
        axis: NabuPlane,
    ):
        axis = NabuPlane.from_value(axis)
        if pag:
            assert db is not None, "if paganin defined, db should not be None"
        if slice_index is not None:
            slice_index = slice_index_to_int(slice_index, scan=scan, axis=axis)

        assert type(db) in (int, type(None))
        if isinstance(scan, NXtomoScan):
            basename, _ = os.path.splitext(scan.master_file)
            basename = os.path.basename(basename)
            try:
                with open_hdf5(scan.master_file) as h5f:
                    if len(h5f.keys()) > 1:
                        # if there is more than one entry in the file append the entry name to the file basename
                        basename = "_".join((basename, scan.entry.lstrip("/")))
            except Exception:
                pass
        else:
            basename = os.path.basename(scan.path)
        if slice_index is None:
            if pag:
                return "_".join((basename + "pag", "db" + str(db).zfill(4)))
            elif ctf:
                return "_".join((basename + "ctf", "db" + str(db).zfill(4)))
            else:
                return basename
        else:
            if pag:
                return "_".join(
                    (
                        basename + "slice_pag",
                        str(slice_index).zfill(6),
                        "db" + str(db).zfill(4),
                        "plane",
                        axis.value,
                    )
                )
            elif ctf:
                return "_".join(
                    (
                        basename + "slice_ctf",
                        str(slice_index).zfill(6),
                        "db" + str(db).zfill(4),
                        "plane",
                        axis.value,
                    )
                )
            else:
                return "_".join(
                    (
                        basename + "slice",
                        str(slice_index).zfill(6),
                        "plane",
                        axis.value,
                    )
                )

    @docstring(_NabuBaseReconstructor)
    def _get_file_basename_reconstruction(self, pag, db, ctf, axis):
        """

        :param pag: is it a paganin reconstruction
        :param db: delta / beta parameter
        :return: basename of the file reconstructed (without any extension)
        """
        return self.get_file_basename_reconstruction(
            scan=self.scan,
            db=db,
            pag=pag,
            slice_index=self.slice_index,
            ctf=ctf,
            axis=axis,
        )


def run_single_slice_reconstruction(
    scan,
    nabu_config,
    dry_run,
    slice_index: int | str | None,
    process_id: int | None = None,
    cluster_config: dict | None = None,
    add_to_latest_reconstructions=True,
    instantiate_class_only=False,
    axis: NabuPlane = NabuPlane.XY,
) -> ResultsRun | None:
    """
    # TODO: might need something like a context or an option "keep" slice in memory

    :param scan:
    :param nabu_config: configuration of nabu process
    :param cluster_config: configuration of cluster (slurm-cluster only for now)
    :param dry_run:
    :param slice_index: slice index to reconstruct.
                                             If str should be "middle"
    :param local:
    :param stdout: file to redirect stdout
    :param stderr: file to redirect stderr
    :param add_to_latest_reconstructions: if true add reconstructed slice to the latest reconstruction.
                                               We wan't to avoid this treatment for saaxis and sadeltebeta for example
    :param instantiate_class_only: if we don't want to run the SingleSliceRunner but only return them. Use case: we want to keep a hand on processing and it can be cancelled
    :return: result of the slice reconstruction if succeed to launch it.
    """
    # TODO: remove local from the function signature
    target = Target.SLURM if cluster_config not in ({}, None) else Target.LOCAL
    axis = NabuPlane.from_value(axis)
    # FIXEME: nabu fails if outer_circle activated and if the axis != z
    if axis != NabuPlane.XY and nabu_config.get("reconstruction", {}).get(
        "clip_outer_circle", False
    ):
        nabu_config["reconstruction"]["clip_outer_circle"] = False

    if process_id is not None:
        try:
            process_name = ProcessManager().get_process(process_id=process_id).name
        except KeyError:
            process_name = "unknow"
    else:
        process_name = ""

    slice_reconstructor = SingleSliceRunner(
        scan=scan,
        config=nabu_config,
        dry_run=dry_run,
        slice_index=slice_index,
        axis=axis,
        target=target,
        cluster_config=cluster_config,
        add_to_latest_reconstructions=add_to_latest_reconstructions,
        process_name=process_name,
    )
    if instantiate_class_only:
        return slice_reconstructor

    try:
        results = slice_reconstructor.run()
    except TimeoutError as e:
        _logger.error(e)
        return None
    else:
        assert len(results) == 1, "only one slice should be reconstructed"
        return results[0]


class NabuSliceMode(_Enum):
    MIDDLE = "middle"
    OTHER = "other"

    @staticmethod
    def getSlices(slices, scan, axis=NabuPlane.XY) -> tuple:
        res = []
        try:
            mode = NabuSliceMode(slices)
        except ValueError:
            try:
                res = utils.retrieve_lst_of_value_from_str(slices, type_=int)
            except Exception:
                pass
        else:
            if mode == mode.MIDDLE:
                axis = NabuPlane(axis)
                if axis is NabuPlane.XY:
                    n_slice = scan.dim_2 or 2048
                elif axis in (NabuPlane.YZ, NabuPlane.XZ):
                    n_slice = scan.dim_1 or 2048
                else:
                    raise NotImplementedError(f"unknow axis {axis}")
                res.append(n_slice // 2)
            else:
                raise ValueError(
                    "there should be only two ways of defining "
                    "slices: middle one or other, by giving "
                    "an unique value or a list or a tuple"
                )
        return tuple(res)


class NabuSlices(NabuSlicesTask):
    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        deprecated_warning(
            name="tomwer.core.process.reconstruction.nabu.nabuslices.Nabuslices",
            type_="class",
            reason="improve readibility",
            since_version="1.2",
            replacement="NabuSlicesTask",
        )
        super().__init__(process_id, varinfo, inputs, node_id, node_attrs, execinfo)



from silx.gui import qt


class SliderPlayWidgetAction(qt.QWidgetAction):

    sigValueChanged = qt.Signal(int)

    def __init__(
        self,
        parent: qt.QWidget | None = None,
        label: str | None = None,
        tooltip: str | None = None,
    ):
        super().__init__(parent)
        self._build(label=label, tooltip=tooltip)

    def _build(self, label: str, tooltip: str):
        widget = qt.QWidget()
        layout = qt.QHBoxLayout()
        widget.setLayout(layout)
        self._spinbox = qt.QSpinBox()
        self._spinbox.setToolTip(tooltip)
        self._spinbox.setRange(1, 1000000)
        self._spinbox.valueChanged.connect(self.sigValueChanged)
        label = qt.QLabel(label)
        label.setToolTip(tooltip)
        layout.addWidget(label)
        layout.addWidget(self._spinbox)
        self.setDefaultWidget(widget)

    def value(self) -> int:
        return self._spinbox.value()

    def setValue(self, value: int):
        self._spinbox.setValue(value)


class PlayButtonContextMenu(qt.QMenu):

    sigFrameRateChanged = qt.Signal(int)

    def __init__(self, parent: qt.QWidget | None = None):
        super().__init__(parent)
        self._build()

    def _build(self):
        self._framerateAction = SliderPlayWidgetAction(
            self, label="FPS:", tooltip="Display speed in frames per second"
        )
        self._framerateAction.sigValueChanged.connect(self.sigFrameRateChanged)
        self._stepAction = SliderPlayWidgetAction(
            self, label="Step:", tooltip="Step between displayed frames"
        )
        self.addAction(self._framerateAction)
        self._framerateAction.setValue(10)
        self.addAction(self._stepAction)
        self._stepAction.setValue(1)

    def getFrameRate(self) -> int:
        return self._framerateAction.value()

    def setFrameRate(self, rate: int):
        self._framerateAction.setValue(rate)

    def getStep(self) -> int:
        return self._stepAction.value()

    def setStep(self, interval: int):
        self._stepAction.setValue(interval)



# coding: utf-8
from __future__ import annotations

import gc
import os
import pickle
import tempfile

from orangecanvas.scheme.readwrite import literal_dumps
from silx.gui import qt
from silx.gui.utils.testutils import SignalListener, TestCaseQt

from orangecontrib.tomwer.widgets.control.DataDiscoveryOW import DataDiscoveryOW
from tomwer.core.scan.scantype import ScanType
from tomwer.core.utils.scanutils import MockBlissAcquisition
from tomoscan.esrf.scan.mock import MockNXtomo, MockEDF


class TestDataDiscovery(TestCaseQt):
    WAIT_TIME = 5000  # thread processing wait time in ms

    def setUp(self):
        super().setUp()
        self.widget = DataDiscoveryOW()
        self.widget.setConfiguration(self.default_settings)
        self.signal_listener = SignalListener()
        self.widget.getProcessingThread().sigDataFound.connect(self.signal_listener)

    def tearDown(self):
        self.widget.getProcessingThread().sigDataFound.disconnect(self.signal_listener)
        self.signal_listener = None
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        gc.collect()

    @property
    def default_settings(self):
        return {
            "start_folder": "/my/folder",
            "file_filter": "*.h5",
            "scan_type_searched": ScanType.NX_TOMO,
        }

    def testSerialization(self):
        """make sure pickling works for the settings"""
        pickle.dumps(self.widget.getConfiguration())

    def testLiteralDumps(self):
        """make sure literal dumps (from orangecanvas) works for the settings"""
        literal_dumps(self.widget.getConfiguration())

    def testProcessingNXtomo(self):
        """Test the widget can find NXtomos"""
        n_scan = 3
        with tempfile.TemporaryDirectory() as root_dir:
            for i_scan in range(n_scan):
                MockNXtomo(
                    scan_path=os.path.join(root_dir, f"sub_folder_{i_scan}", "scan"),
                    n_proj=10,
                    n_ini_proj=10,
                    dim=10,
                )
            self.widget.setFolderObserved(root_dir)
            self.widget.start_discovery(wait=self.WAIT_TIME)
            while self.qapp.hasPendingEvents():
                self.qapp.processEvents()

            assert self.signal_listener.callCount() == n_scan

    def testProcessingBlissRawScan(self):
        """Test the widget can find bliss raw scans"""
        with tempfile.TemporaryDirectory() as root_dir:
            n_scan = 2
            for i_scan in range(n_scan):
                MockBlissAcquisition(
                    n_sample=1,
                    n_sequence=1,
                    n_scan_per_sequence=1,
                    n_darks=1,
                    n_flats=1,
                    with_nx=True,
                    output_dir=os.path.join(root_dir, f"sub_folder_{i_scan}", "scan"),
                )

            self.widget.setFilePattern("*.edf")
            self.widget.setSearchScanType(ScanType.BLISS)
            self.widget.setFolderObserved(root_dir)
            self.widget.start_discovery(wait=self.WAIT_TIME)
            while self.qapp.hasPendingEvents():
                self.qapp.processEvents()

            assert (
                self.signal_listener.callCount() == 0
            )  # because of the filtering on files
            self.widget.setFilePattern(None)
            self.widget.start_discovery(wait=self.WAIT_TIME)
            while self.qapp.hasPendingEvents():
                self.qapp.processEvents()

            assert self.signal_listener.callCount() == n_scan

    def testProcessingSpecScan(self):
        """Test the widget can find bliss raw scans"""
        with tempfile.TemporaryDirectory() as root_dir:
            n_scan = 4
            for i_scan in range(n_scan):
                MockEDF(
                    n_radio=4,
                    n_ini_radio=4,
                    scan_path=os.path.join(root_dir, f"sub_folder_{i_scan}", "scan"),
                )

            self.widget.setSearchScanType(ScanType.BLISS)
            self.widget.setFilePattern(None)
            self.widget.setFolderObserved(root_dir)
            self.widget.start_discovery(wait=self.WAIT_TIME)
            while self.qapp.hasPendingEvents():
                self.qapp.processEvents()

            assert (
                self.signal_listener.callCount() == 0
            )  # because of the scan type searched
            self.widget.setSearchScanType(ScanType.SPEC)
            self.widget.setFilePattern(None)
            self.widget.start_discovery(wait=self.WAIT_TIME)
            while self.qapp.hasPendingEvents():
                self.qapp.processEvents()
            assert self.signal_listener.callCount() == n_scan



# coding: utf-8

"""
This module is used to manage observations. Initially on files.
Observations are runned on a thread and run each n seconds.
They are manage by thread and signals
"""

from time import time


class Timer:
    def __init__(self, title):
        self.title = title

    def __enter__(self):
        self._start_time = time()

    def __exit__(self, type, value, traceback):
        exec_time = time() - self._start_time
        print(f"{self.title} exec time: {exec_time} s")



from __future__ import annotations

import logging

from silx.gui import qt

from tomwer.core.process.reconstruction.axis import mode as axis_mode
from tomwer.core.process.reconstruction.axis.anglemode import CorAngleMode
from tomwer.gui.utils.scrollarea import QComboBoxIgnoreWheel
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.utils.qt_utils import block_signals
from tomwer.synctools.axis import QAxisRP
from .ManualFramesSelection import ManualFramesSelection

_logger = logging.getLogger(__name__)


class InputWidget(qt.QWidget):
    """
    Widget used to define the radios or the sinogram to be used for computing the cor
    Used as a tab of the AxisSettingsTabWidget
    """

    sigChanged = qt.Signal()
    """Signal emitted when input changed"""

    _sigUrlChanged = qt.Signal()
    """Signal emit when url to be used changed"""

    def __init__(self, parent=None, axis_params=None):
        assert isinstance(axis_params, QAxisRP)
        self._blockUpdateAxisParams = False
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())

        # radio input
        self._radioGB = qt.QGroupBox(self)
        self._radioGB.setTitle("radios")
        self._radioGB.setLayout(qt.QVBoxLayout())
        self._radioGB.setCheckable(True)
        self.layout().addWidget(self._radioGB)
        ## angle mode
        self._angleModeWidget = _AngleSelectionWidget(
            parent=self, axis_params=axis_params
        )
        self._radioGB.layout().addWidget(self._angleModeWidget)
        self._axis_params = axis_params

        # sinogram input
        self._sinogramGB = qt.QGroupBox(self)
        self._sinogramGB.setLayout(qt.QFormLayout())

        self._sinogramGB.setTitle("sinogram")
        self._sinogramGB.setCheckable(True)
        self.layout().addWidget(self._sinogramGB)
        ##  sinogram line
        self._sinogramLineSB = _SliceSelector(self)
        self._sinogramGB.layout().addRow("line", self._sinogramLineSB)
        ##  sinogram subsampling
        self._sinogramSubsampling = qt.QSpinBox(self)
        self._sinogramSubsampling.setRange(1, 1000)
        self._sinogramSubsampling.setValue(10)
        self._sinogramGB.layout().addRow("subsampling", self._sinogramSubsampling)

        self._spacer = qt.QWidget(self)
        self._spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(self._spacer)

        # set up
        self._sinogramGB.setChecked(False)

        # connect signal / slot
        self._sinogramGB.toggled.connect(self._sinogramChecked)
        self._radioGB.toggled.connect(self._radiosChecked)
        self._sinogramSubsampling.valueChanged.connect(self._changed)
        self._sinogramLineSB.sigChanged.connect(self._changed)
        self._angleModeWidget.sigChanged.connect(self._sigUrlChanged)

    def setScan(self, scan: TomwerScanBase):
        if scan is not None:
            self._angleModeWidget.setScan(scan)
            self._angleModeWidget.setScanRange(scan.scan_range)

    def setAxisParams(self, axis_params):
        with block_signals(axis_params):
            with block_signals(self._axis_params):
                self._blockUpdateAxisParams = True

                if axis_params is not None:
                    assert isinstance(axis_params, QAxisRP)
                    with block_signals(self._sinogramGB):
                        self._sinogramChecked(
                            axis_params.mode.requires_sinogram_index(), on_load=True
                        )
                    self._sinogramLineSB.setSlice(axis_params.sinogram_line)
                    self._sinogramSubsampling.setValue(axis_params.sinogram_subsampling)
                self._angleModeWidget.setAxisParams(axis_params)
                self._axis_params = axis_params

        self._blockUpdateAxisParams = False

    def getSinogramLine(self) -> str | int:
        return self._sinogramLineSB.getSlice()

    def getSinogramSubsampling(self) -> int:
        return self._sinogramSubsampling.value()

    def _sinogramChecked(self, checked, on_load=False):
        with block_signals(self._radioGB):
            with block_signals(self._sinogramGB):
                if checked:
                    self._radioGB.setChecked(False)
                    self._sinogramGB.setChecked(True)
                elif self._radioGB.isEnabled():
                    self._radioGB.setChecked(not checked)
                elif on_load:
                    self._sinogramGB.setChecked(checked)
                else:
                    # ignore it if radio disabled
                    self._sinogramGB.setChecked(True)
        self._changed()

    def _radiosChecked(self, checked, on_load=False):
        with block_signals(self._radioGB):
            with block_signals(self._sinogramGB):
                if checked:
                    self._sinogramGB.setChecked(False)
                    self._radioGB.setChecked(True)
                elif self._sinogramGB.isEnabled():
                    self._sinogramGB.setChecked(not checked)
                elif on_load:
                    self._radioGB.setChecked(checked)
                else:
                    # ignore it if sinogram disabled
                    self._radioGB.setChecked(True)
        self._changed()

    def _changed(self, *args, **kwargs):
        self._updateAxisParams()
        self.sigChanged.emit()

    def _updateAxisParams(self):
        if not self._blockUpdateAxisParams:
            self._axis_params.sinogram_line = self.getSinogramLine()
            self._axis_params.sinogram_subsampling = self.getSinogramSubsampling()

    def setValidInputs(self, modes: list | tuple):
        """
        Define possible inputs.

        :raises: ValueError if modes are invalid
        """
        modes = set(modes)
        for mode in modes:
            try:
                axis_mode._InputType(mode)
            except ValueError:
                raise ValueError(
                    f"mode {mode} should be an instance of {axis_mode._InputType}"
                )
        if len(modes) == 2:
            self._sinogramGB.setEnabled(True)
            self._radioGB.setEnabled(True)
        elif len(modes) > 2:
            raise ValueError(f"invalid input {modes}")
        elif len(modes) < 0:
            raise ValueError("modes is empty")
        else:
            mode = axis_mode._InputType(modes.pop())
            if mode is axis_mode._InputType.SINOGRAM:
                self._sinogramGB.setEnabled(True)
                self._radioGB.setEnabled(False)
                self._sinogramGB.setChecked(True)
            elif mode is axis_mode._InputType.RADIOS_X2:
                self._radioGB.setEnabled(True)
                self._sinogramGB.setEnabled(False)
                self._radioGB.setChecked(True)
            elif mode is axis_mode._InputType.COMPOSITE:
                # those mode are neither sinogram neither radio. Now one of the two will be checked but without any much meaning
                self._radioGB.setEnabled(False)
                self._sinogramGB.setEnabled(False)
            else:
                raise ValueError(f"Nothing implemented for {mode.value}")


class _AngleSelectionWidget(qt.QWidget):
    """Group box to select the angle to used for cor calculation
    (0-180, 90-270 or manual)"""

    sigChanged = qt.Signal()
    """signal emitted when the selected angle changed"""

    def __init__(self, parent=None, axis_params=None):
        assert isinstance(axis_params, QAxisRP)
        qt.QWidget.__init__(
            self,
            parent=parent,
        )
        self.setLayout(qt.QVBoxLayout())
        self._groupBoxMode = qt.QGroupBox(
            self, title="Angles to use for axis calculation"
        )
        self._groupBoxMode.setLayout(qt.QHBoxLayout())
        self.layout().addWidget(self._groupBoxMode)

        self._corButtonsGps = qt.QButtonGroup(parent=self)
        self._corButtonsGps.setExclusive(True)
        self._qrbCOR_0_180 = qt.QRadioButton("0-180", parent=self)
        self._groupBoxMode.layout().addWidget(self._qrbCOR_0_180)
        self._qrbCOR_90_270 = qt.QRadioButton("90-270", parent=self)
        self._qrbCOR_90_270.setToolTip(
            "pick radio closest to angles 90° and "
            "270°. If disable mean that the scan "
            "range is 180°"
        )
        self._groupBoxMode.layout().addWidget(self._qrbCOR_90_270)
        self._qrbCOR_manual = qt.QRadioButton("other", parent=self)
        self._qrbCOR_manual.setVisible(True)
        self._groupBoxMode.layout().addWidget(self._qrbCOR_manual)
        # add all button to the button group
        for b in (self._qrbCOR_0_180, self._qrbCOR_90_270, self._qrbCOR_manual):
            self._corButtonsGps.addButton(b)

        self.setAxisParams(axis_params)

        self._manualFrameSelection = ManualFramesSelection(self)
        self.layout().addWidget(self._manualFrameSelection)
        self._manualFrameSelection.setVisible(False)

        # connect signal / Slot
        self._corButtonsGps.buttonClicked.connect(self._angleModeChanged)
        self._manualFrameSelection.sigChanged.connect(self._changed)

    def setScan(self, scan: TomwerScanBase):
        if scan is not None:
            self.setScanRange(scan.scan_range)
        self._manualFrameSelection.setScan(scan=scan)

    def setScanRange(self, scanRange):
        if scanRange == 180:
            self._qrbCOR_90_270.setEnabled(False)
            # force using 0-180 if was using 90-270
            if self._qrbCOR_90_270.isChecked():
                self._qrbCOR_0_180.setChecked(True)
                self._axis_params.angle_mode = CorAngleMode.use_0_180
        else:
            self._qrbCOR_90_270.setEnabled(True)

    def setAngleMode(self, mode):
        """

        :param mode: mode to use (can be manual , 90-270 or 0-180)
        """
        assert isinstance(mode, CorAngleMode)
        if mode == CorAngleMode.use_0_180:
            self._qrbCOR_0_180.setChecked(True)
        elif mode == CorAngleMode.use_90_270:
            self._qrbCOR_90_270.setChecked(True)
        else:
            self._qrbCOR_manual.setChecked(True)

    def getAngleMode(self) -> CorAngleMode:
        """

        :return: the angle to use for the axis calculation
        """
        if self._qrbCOR_90_270.isChecked():
            return CorAngleMode.use_90_270
        elif self._qrbCOR_0_180.isChecked():
            return CorAngleMode.use_0_180
        else:
            return CorAngleMode.manual_selection

    def setAxisParams(self, axis_params):
        with block_signals(self):
            self._axis_params = axis_params
            # set up
            self.setAngleMode(axis_params.angle_mode)

    def _angleModeChanged(self, *args, **kwargs):
        self._axis_params.angle_mode = self.getAngleMode()
        if self.getAngleMode() is CorAngleMode.manual_selection:
            self._axis_params.angle_mode_extra = (
                self._manualFrameSelection.getFramesUrl()
            )
        else:
            self._axis_params.angle_mode_extra = None
        self._manualFrameSelection.setVisible(
            self.getAngleMode() is CorAngleMode.manual_selection
        )
        self._changed()

    def _changed(self):
        self.sigChanged.emit()


class _SliceSelector(qt.QWidget):
    sigChanged = qt.Signal()
    """signal emit when the selected slice change"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        self.setContentsMargins(0, 0, 0, 0)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self._modeCB = QComboBoxIgnoreWheel(self)
        self._modeCB.addItem("middle")
        self._modeCB.addItem("other")
        self.layout().addWidget(self._modeCB)
        self._otherSB = qt.QSpinBox(self)
        self._otherSB.setRange(0, 10000)
        self.layout().addWidget(self._otherSB)

        # connect signal / slot
        self._otherSB.valueChanged.connect(self._valueChanged)
        self._modeCB.currentIndexChanged.connect(self._modeChanged)
        # set up
        self._modeChanged()

    def getSlice(self) -> int | str:
        "return a specific slice index or 'middle'"
        if self.getMode() == "middle":
            return "middle"
        else:
            return self._otherSB.value()

    def setSlice(self, slice_):
        if slice_ is None:
            return
        if slice_ == "middle":
            idx = self._modeCB.findText("middle")
            self._modeCB.setCurrentIndex(idx)
        else:
            idx = self._modeCB.findText("other")
            self._modeCB.setCurrentIndex(idx)
            self._otherSB.setValue(slice_)
        self.sigChanged.emit()

    def getMode(self):
        return self._modeCB.currentText()

    def _valueChanged(self):
        self.sigChanged.emit()

    def _modeChanged(self, *args, **kwargs):
        self._otherSB.setVisible(self.getMode() == "other")
        self._valueChanged()



"""contain the SAAxisProcess. Half automatic center of rotation calculation"""

from __future__ import annotations

import logging
import os
import h5py

import numpy
from multiprocessing import Pool
from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from silx.io.url import DataUrl
from silx.io.utils import h5py_read_dataset

from tomwer.core.utils.deprecation import deprecated_warning
from tomoscan.io import HDF5File

import tomwer.version
from tomwer.core.process.reconstruction.axis import AxisRP
from tomwer.core.process.reconstruction.nabu.nabuscores import (
    run_nabu_multicor,
)
from tomwer.core.process.reconstruction.nabu.nabuslices import (
    interpret_tomwer_configuration,
)
from tomwer.core.process.reconstruction.scores import (
    ComputedScore,
    apply_roi,
    compute_score,
    get_disk_mask_radius,
)
from tomwer.core.process.reconstruction.scores.params import ScoreMethod
from tomwer.core.process.reconstruction.utils.cor import relative_pos_to_absolute
from tomwer.core.process.task import Task
from tomwer.core.process import utils as core_utils
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils import logconfig
from tomwer.core.utils.scanutils import data_identifier_to_scan
from tomwer.core.utils.slurm import is_slurm_available
from tomwer.core.process.reconstruction.nabu.utils import update_nabu_config_for_tiff_3d

from tomwer.io.utils.h5pyutils import EntryReader
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.io.utils.utils import get_slice_data
from tomwer.io.utils import format_stderr_stdout
from tomwer.core.process.reconstruction.nabu.nabucommon import (
    ResultsLocalRun,
    ResultSlurmRun,
    ResultsWithStd,
)
from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.process.reconstruction.saaxis.params import ReconstructionMode
from tomwer.core.process.reconstruction.nabu.utils import (
    get_multi_cor_recons_volume_identifiers,
    get_nabu_multicor_file_prefix,
)

from .params import SAAxisParams

_logger = logging.getLogger(__name__)


DEFAULT_RECONS_FOLDER = "multi_cor_results"


def one_slice_several_cor(
    scan,
    configuration: dict,
    process_id: int | None = None,
) -> tuple:
    """
    Run a slice reconstruction using nabu per Center Of Rotation (cor) provided
    Then for each compute a score (quality) of the center of rotation

    .. warning:: if target if the slurm cluster this will wait for the processing to be done to return the result.
                 as this function is returning the result of the score process on reconstructed slices

    :param scan:
    :param configuration: nabu reconstruction parameters (can include 'slurm-cluster' key defining the slurm configuration)
    :param process_id: process id
    :return: cor_reconstructions, outs, errs
             cor_reconstructions is a dictionary of cor as key and a tuple
             (url, score) as value
    """
    task = SAAxisTask(
        process_id=process_id,
        inputs={
            "data": scan,
            "sa_axis_params": configuration,
            "serialize_output_data": False,
        },
    )
    task.run()
    return (
        task.outputs.scores,
        task.outputs.std_out,
        task.outputs.std_err,
        task.outputs.rois,
    )


class SAAxisTask(
    Task,
    SuperviseProcess,
    input_names=("data", "sa_axis_params"),
    output_names=("data", "best_cor"),
    optional_input_names=(
        "dry_run",
        "dump_roi",
        "dump_process",
        "serialize_output_data",
        "compute_scores",  # for GUI we want to post pone the score calculation
        "pool_size",
    ),
):
    """
    Main process to launch several reconstruction of a single slice with
    several Center Of Rotation (cor) values

    As the saaxis is integrating the score calculation we will never get a future_tomo_scan as output
    """

    DEFAULT_POOL_SIZE = 10

    def __init__(
        self, process_id=None, inputs=None, varinfo=None, node_attrs=None, execinfo=None
    ):
        Task.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        SuperviseProcess.__init__(self, process_id=process_id)
        self._dry_run = inputs.get("dry_run", False)
        self._dump_process = inputs.get("dump_process", True)
        self._dump_roi = inputs.get("dump_roi", False)
        self._std_outs = tuple()
        self._std_errs = tuple()
        self._current_processing = None
        self._cancelled = False

    @property
    def std_outs(self):
        return self._std_outs

    @property
    def std_errs(self):
        return self._std_errs

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    @property
    def dry_run(self):
        return self._dry_run

    @property
    def dump_roi(self):
        return self._dump_roi

    @dump_roi.setter
    def dump_roi(self, dump):
        self._dump_roi = dump

    @staticmethod
    def autofocus(scan) -> float | None:
        scores = scan.saaxis_params.scores
        if scores is None:
            return
        score_method = scan.saaxis_params.score_method
        best_cor, best_score = None, 0
        for cor, (_, score_cls) in scores.items():
            if score_cls is None:  # if score calculation failed
                continue
            score = score_cls.get(score_method)
            if score is None:
                continue
            if score > best_score:
                best_cor, best_score = cor, score
        scan.saaxis_params.autofocus = best_cor
        if scan.axis_params is None:
            # create parameter if needed because will set it once he find the best cor
            scan.axis_params = AxisRP()
        scan.axis_params.frame_width = scan.dim_1
        scan.axis_params.set_relative_value(best_cor)
        return best_cor

    def _config_preprocessing(
        self, scan, config, file_format, output_dir, cluster_config
    ):
        """convert general configuration to nabu - single reconstruction - configuration"""
        nabu_configurations = interpret_tomwer_configuration(config, scan=None)
        if len(nabu_configurations) == 0:
            raise RuntimeWarning(
                "Unable to get a valid nabu configuration for " "reconstruction."
            )
        elif len(nabu_configurations) > 1:
            _logger.warning(
                "Several configuration found for nabu (you probably "
                "ask for several delta/beta value or several slices). "
                "Picking the first one."
            )

        # work on file name...
        if output_dir is None:
            output_dir = os.path.join(scan.path, DEFAULT_RECONS_FOLDER)

        nabu_configuration = nabu_configurations[0][0]
        if cluster_config == {}:
            cluster_config = None
        is_cluster_job = cluster_config is not None
        if is_cluster_job and not is_slurm_available():
            raise ValueError(
                "job on cluster requested but no access to slurm cluster found"
            )

        # handle reconstruction section
        if "reconstruction" not in nabu_configuration:
            nabu_configuration["reconstruction"] = {}
        nabu_configuration["reconstruction"]["rotation_axis_position"] = ""
        # handle output section
        if "output" not in nabu_configuration:
            nabu_configuration["output"] = {}
        nabu_configuration["output"]["location"] = output_dir
        nabu_configuration["output"]["file_format"] = file_format
        update_nabu_config_for_tiff_3d(nabu_configuration)

        # handle resources section
        nabu_configuration["resources"] = nabu_configuration.get("resources", {})
        nabu_configuration["resources"]["method"] = "local"

        return nabu_configuration

    def _run_nabu_multicor(
        self,
        scan,
        nabu_config,
        cors,
        slice_index,
        file_format,
        cluster_config: dict | None,
        dry_run=False,
    ):
        if not (cluster_config is None or isinstance(cluster_config, dict)):
            raise TypeError(
                f"cluster_config is expected to be a dict. Get {type(cluster_config)} instead."
            )
        runner = run_nabu_multicor(
            nabu_config=nabu_config,
            scan=scan,
            cors=cors,
            slice_index=slice_index,
            dry_run=dry_run,
            file_format=file_format,
            cluster_config=cluster_config if cluster_config is not None else None,
            process_id=self.process_id,
            instantiate_classes_only=True,
            output_file_prefix_pattern="cor_{file_name}_{value}",  # as the cor is evolving, create different files to make sure the name will be unique
        )

        future_tomo_obj = None
        recons_urls = dict()
        std_outs = []
        std_errs = []

        self._current_processing = runner
        try:
            result = runner.run()
        except TimeoutError as e:
            _logger.error(e)
        else:
            success = result.success
            if isinstance(result, ResultsWithStd):
                std_outs.append(result.std_out)
                std_errs.append(result.std_err)
            if isinstance(result, ResultsLocalRun):
                recons_urls = {
                    cor: recons for cor, recons in zip(cors, result.results_identifiers)
                }
            if isinstance(result, ResultSlurmRun):
                future_tomo_obj = FutureTomwerObject(
                    tomo_obj=scan,
                    process_requester_id=self.process_id,
                    futures=result.future_slurm_jobs,
                )
        return success, recons_urls, (future_tomo_obj,), std_outs, std_errs

    def _resolve_futures(
        self,
        scan,
        nabu_config,
        slice_index,
        file_format,
        cors,
        cor_reconstructions,
        future_tomo_objs: dict,
        output_dir,
    ):
        """
        in case the task is launching jobs over slurm wait for them to be finished before resuming 'standard processing'
        """
        if output_dir is None:
            output_dir = os.path.join(scan.path, DEFAULT_RECONS_FOLDER)

        file_prefix = get_nabu_multicor_file_prefix(scan)

        for future_tomo_obj in future_tomo_objs:
            if self._cancelled:
                break

            if future_tomo_obj is None:
                continue

            future_tomo_obj.results()
            if future_tomo_obj.cancelled() or future_tomo_obj.exceptions():
                continue

            for cor in cors:
                cor_nabu_ref = relative_pos_to_absolute(
                    relative_pos=cor,
                    det_width=scan.dim_1,
                )
                volume_identifiers = get_multi_cor_recons_volume_identifiers(
                    scan=scan,
                    slice_index=slice_index,
                    location=nabu_config["output"]["location"],
                    file_prefix=file_prefix,
                    file_format=file_format,
                    cors=(cor_nabu_ref,),
                )
                volume_identifier = volume_identifiers.get(cor_nabu_ref, None)
                if volume_identifier is None:
                    _logger.warning(
                        f"failed to load volume for {cor}. Something went wrong on slurm submission job"
                    )
                cor_reconstructions[cor] = volume_identifier

    def _post_processing(self, scan, slice_index, cor_reconstructions: dict):
        """
        compute score along the different slices

        :param cor_reconstructions: key is expected to be a float with the cor value and the value is expected to be a volume identifier (volume with a single frame)
        """
        post_processing = _PostProcessing(
            slice_index=slice_index,
            scan=scan,
            cor_reconstructions=cor_reconstructions,
            pool_size=self.get_input_value("pool_size", self.DEFAULT_POOL_SIZE),
        )
        post_processing._cancelled = self._cancelled
        self._current_processing = post_processing
        return post_processing.run()

    def _compute_mess_details(self, mess=""):
        """
        util to join a message and nabu std err and std out
        """
        nabu_logs = []
        for std_err, std_out in zip(self._std_errs, self.std_outs):
            nabu_logs.append(format_stderr_stdout(stdout=std_out, stderr=std_err))
        self._nabu_log = nabu_logs
        nabu_logs.insert(0, mess)
        return "\n".join(nabu_logs)

    @staticmethod
    def _preprocess_slice_index(slice_index, mode: ReconstructionMode):
        if isinstance(slice_index, str):
            if not slice_index == "middle":
                raise ValueError(f"slice index {slice_index} not recognized")
            else:
                return slice_index
        elif not len(slice_index) == 1:
            raise ValueError(f"{mode.value} mode only manage one slice")
        else:
            return list(slice_index.values())[0]

    def get_output_dir(self, params: SAAxisParams, scan: TomwerScanBase):
        output_dir = params.output_dir or None
        if output_dir is None:
            output_dir = (
                params.nabu_recons_params.get("output", {}).get("location", None)
                or None
            )
            if output_dir is None:
                output_dir = os.path.join(scan.path, DEFAULT_RECONS_FOLDER)
        return output_dir

    def run(self):
        scan = data_identifier_to_scan(self.inputs.data)
        if scan is None:
            self.outputs.data = None
            return
        if isinstance(scan, TomwerScanBase):
            scan = scan
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise ValueError(f"input type of {scan}: {type(scan)} is not managed")
        # TODO: look and update if there is some nabu reconstruction
        # or axis information to be used back
        configuration = self.inputs.sa_axis_params
        params = SAAxisParams.from_dict(configuration)
        # insure output dir is created
        params.output_dir = self.get_output_dir(params=params, scan=scan)
        if not os.path.exists(params.output_dir):
            os.makedirs(params.output_dir)

        # try to find an estimated cor
        #  from a previously computed cor
        if params.estimated_cor is None and scan.axis_params is not None:
            relative_cor = scan.axis_params.relative_cor_value
            if relative_cor is not None and numpy.issubdtype(
                type(relative_cor), numpy.number
            ):
                params.estimated_cor = relative_cor
                _logger.info(
                    f"{scan}: set estimated cor from previously computed cor ({params.estimated_cor})"
                )
        #  from scan.x_rotation_axis_pixel_position
        if (
            params.estimated_cor is None
            and scan.x_rotation_axis_pixel_position is not None
        ):
            params.estimated_cor = scan.x_rotation_axis_pixel_position
            _logger.info(
                f"{scan}: set estimated cor from motor position ({params.estimated_cor})"
            )
        if scan.dim_1 is not None:
            params.image_width = scan.dim_1
        scan.saaxis_params = params

        mode = ReconstructionMode(params.mode)
        if mode is not ReconstructionMode.VERTICAL:
            raise ValueError(f"{mode} is not handled for now")

        nabu_config = configuration.get("nabu_params", {})
        nabu_output_config = configuration.get("output", {})
        file_format = nabu_output_config.get("file_format", "hdf5")
        slice_index = self._preprocess_slice_index(
            params.slice_indexes,
            mode=mode,
        )
        cluster_config = params.cluster_config
        dry_run = self._dry_run

        # step one: complete nabu configuration(s)
        nabu_config = self._config_preprocessing(
            scan=scan,
            config=nabu_config,
            file_format=file_format,
            output_dir=params.output_dir,
            cluster_config=cluster_config,
        )
        # step 2: run reconstructions
        cors_res = {}
        rois = {}

        try:
            (
                _,
                cor_reconstructions,
                future_tomo_objs,
                self._std_outs,
                self._std_errs,
            ) = self._run_nabu_multicor(
                scan=scan,
                nabu_config=nabu_config,
                cors=tuple(params.cors),
                slice_index=slice_index,
                file_format=file_format,
                cluster_config=cluster_config,
                dry_run=dry_run,
            )
        except Exception as e:
            _logger.error(e)
            mess = f"sa-axis -nabu- computation for {str(scan)} failed."
            state = DatasetState.FAILED
        else:
            # step 3: wait for future if any
            self._resolve_futures(
                scan=scan,
                nabu_config=nabu_config,
                slice_index=slice_index,
                file_format=file_format,
                cor_reconstructions=cor_reconstructions,
                cors=tuple(params.cors),
                future_tomo_objs=future_tomo_objs,
                output_dir=params.output_dir,
            )

            # step 4: run post processing (compute score for each slice)
            if self.get_input_value("compute_scores", True):
                try:
                    cors_res, rois = self._post_processing(
                        scan=scan,
                        slice_index=slice_index,
                        cor_reconstructions=cor_reconstructions,
                    )
                except Exception as e:
                    _logger.error(e)
                    mess = (
                        f"sa-axis -post-processing- computation for {str(scan)} failed."
                    )
                    state = DatasetState.FAILED
                    cors_res = {}
                else:
                    state = DatasetState.WAIT_USER_VALIDATION
                    mess = "sa-axis computation succeeded"
            else:
                cors_res = {}
                state = DatasetState.FAILED
                mess = "couldn't find 'compute_scores'"

        if self._cancelled:
            state = DatasetState.CANCELLED
            mess = "scan cancelled by the user"

        ProcessManager().notify_dataset_state(
            dataset=scan,
            process=self,
            state=state,
            details=self._compute_mess_details(mess),
        )

        scan.saaxis_params.scores = cors_res
        best_relative_cor = self.autofocus(scan=scan)

        if best_relative_cor is not None:
            scan.axis_params.set_relative_value(best_relative_cor)

        self._process_end(scan=scan, cors_res=cors_res, score_rois=rois)

        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan
        self.outputs.best_cor = best_relative_cor

    def _process_end(self, scan, cors_res, score_rois):
        assert isinstance(scan, TomwerScanBase)
        state = ProcessManager().get_dataset_state(
            dataset_id=scan.get_identifier(), process=self
        )
        if state not in (
            DatasetState.CANCELLED,
            DatasetState.FAILED,
            DatasetState.SKIPPED,
        ):
            try:
                extra = {
                    logconfig.DOC_TITLE: self._scheme_title,
                    logconfig.SCAN_ID: str(scan),
                }
                slice_index = self.inputs.sa_axis_params.get("slice_index", None)

                if cors_res is None:
                    info = f"fail to compute cor scores of slice {slice_index} for scan {scan}."
                    _logger.processFailed(info, extra=extra)
                    ProcessManager().notify_dataset_state(
                        dataset=scan,
                        process=self,
                        state=DatasetState.FAILED,
                        details=info,
                    )
                else:
                    info = (
                        f"cor scores of slice {slice_index} for scan {scan} computed."
                    )
                    _logger.processSucceed(info, extra=extra)
                    ProcessManager().notify_dataset_state(
                        dataset=scan,
                        process=self,
                        state=DatasetState.WAIT_USER_VALIDATION,
                        details=info,
                    )
            except Exception as e:
                _logger.error(e)
            else:
                if self._dump_process:
                    self.save_results_to_disk(scan=scan)
                    if self.dump_roi:
                        self.dump_rois(scan, score_rois=score_rois)

    @staticmethod
    def dump_rois(scan, score_rois):
        if scan.saaxis_params.scores in (None, {}):
            return

        process_url = SAAxisTask.get_results_url(scan=scan)

        with HDF5File(process_url.file_path(), mode="w") as h5f:
            nx_process = h5f.require_group(process_url.data_path())
            score_roi_grp = nx_process.require_group("score_roi")
            for cor, roi in score_rois.items():
                score_roi_grp[str(cor)] = roi
                score_roi_grp[str(cor)].attrs["interpretation"] = "image"

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        return "semi-automatic axis"

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return tomwer.version.version

    @staticmethod
    def definition():
        """definition of the process"""
        return "Semi automatic center of rotation / axis calculation"

    @staticmethod
    def get_results_url(scan):
        return DataUrl(
            file_path=scan.get_relative_file(
                os.path.join(DEFAULT_RECONS_FOLDER, "tomwer_saaxis.h5")
            ),
            data_path=scan.entry or "entry",
            scheme="silx",
        )

    @staticmethod
    def save_results_to_disk(scan):
        if scan.saaxis_params.scores in (None, {}):
            return

        saaxis_results_url = SAAxisTask.get_results_url(scan=scan)

        if not os.path.exists(saaxis_results_url.file_path()):
            _logger.error("no result saved")
            return

        # save it to the file
        with HDF5File(saaxis_results_url.file_path(), mode="a") as h5f:
            nx_process = h5f.require_group(saaxis_results_url.data_path())
            if "NX_class" not in nx_process.attrs:
                nx_process.attrs["NX_class"] = "NXprocess"

            results = nx_process.require_group("results")
            for cor, (url, score) in scan.saaxis_params.scores.items():
                results_cor = results.require_group(str(cor))
                for method in ScoreMethod:
                    method_score = score.get(method)
                    if method_score is None:
                        results_cor[method.value] = "None"
                    else:
                        results_cor[method.value] = method_score

                link_path = os.path.relpath(
                    url.file_path(),
                    os.path.dirname(saaxis_results_url.file_path()),
                )
                results_cor["reconstructed_slice"] = h5py.ExternalLink(
                    link_path, url.data_path()
                )

    @staticmethod
    def load_results_from_disk(scan):
        saaxis_results_url = SAAxisTask.get_results_url(scan=scan)

        if saaxis_results_url.file_path() is None or not os.path.exists(
            saaxis_results_url.file_path()
        ):
            _logger.warning(
                "Unable to find process file. Unable to read " "existing processing"
            )
            return None, None

        try:
            with EntryReader(saaxis_results_url) as h5f_entry_node:
                scores = core_utils.get_scores(h5f_entry_node)
                if (
                    "results" in h5f_entry_node
                    and "center_of_rotation" in h5f_entry_node["results"]
                ):
                    selected = h5py_read_dataset(
                        h5f_entry_node["results"]["center_of_rotation"]
                    )
                else:
                    _logger.warning(f"no results found for {scan}")
                    selected = None
                return scores, selected
        except ValueError:
            _logger.warning(f"Data path ({saaxis_results_url.data_path()}) not found")
            return None, None

    def cancel(self):
        """
        stop current processing
        """
        if self._current_processing is not None:
            self._cancelled = True
            self._current_processing.cancel()


class _PostProcessing:
    """class used to run SA-axis post-processing on reconstructed slices"""

    def __init__(self, cor_reconstructions, slice_index, scan, pool_size) -> None:
        self._cor_reconstructions = cor_reconstructions
        self._slice_index = slice_index
        self._scan = scan
        self._cancelled = False
        self.pool_size = pool_size

    @staticmethod
    def compute_score(item: tuple):
        cor, (url, data), mask_disk_radius, cancelled = item
        if cancelled:
            return (None, None), None

        if data is None:
            score = None
            data_roi = None
        else:
            if not isinstance(data, numpy.ndarray):
                raise TypeError(
                    f"data should be a numpy array. Get {type(data)} instead"
                )
            assert data.ndim == 2, f"data should be 2D. Get {data.ndim} instead"
            data_roi = apply_roi(data=data, radius=mask_disk_radius, url=url)

            # move data_roi to [0-1] range
            #  preprocessing: get percentile 0 and 99 from image and
            #  "clean" highest and lowest pixels from it
            min_p, max_p = numpy.percentile(data_roi, (1, 99))
            data_roi_int = data_roi[...]
            data_roi_int[data_roi_int < min_p] = min_p
            data_roi_int[data_roi_int > max_p] = max_p
            data_roi_int = (data_roi_int - min_p) / (max_p - min_p)

            score = ComputedScore(
                tv=compute_score(data=data_roi_int, method=ScoreMethod.TV),
                std=compute_score(data=data_roi_int, method=ScoreMethod.STD),
                tomo_consistency=None,
            )
        return {cor: (url, score)}, {cor: data_roi}

    def run(self):
        datasets = self.load_datasets()
        assert isinstance(datasets, dict)
        mask_disk_radius = get_disk_mask_radius(datasets)
        with Pool(self.pool_size) as pool:
            res = pool.map(
                self.compute_score,
                [
                    (
                        *item,
                        mask_disk_radius,
                        self._cancelled,
                    )
                    for item in datasets.items()
                ],
            )
        scores = {}
        rois = {}
        for mydict in res:
            myscores, myrois = mydict
            scores.update(myscores)
            rois.update(myrois)
        return scores, rois

    @staticmethod
    def _load_dataset(item: tuple):
        cor, volume_identifier = item
        if volume_identifier is None:
            return {cor: (None, None)}

        volume = VolumeFactory.create_tomo_object_from_identifier(volume_identifier)
        urls = tuple(volume.browse_data_urls())
        if len(urls) == 0:
            _logger.error(
                f"volume {volume.get_identifier().to_str()} has no url / slices. Unable to load any data."
            )
            return {cor: (None, None)}
        if len(urls) != 1:
            _logger.error(
                f"volume is expected to have at most one url (single slice volume). get {len(urls)} - most likely nabu reconstruction failed. Do you have GPU ? Are the requested COR values valid ? - Especially for Half-acquisition"
            )
        url = urls[0]
        if not isinstance(url, (DataUrl, str)):
            raise TypeError(f"url is expected to be a str or DataUrl not {type(url)}")

        try:
            data = get_slice_data(url=url)
        except Exception as e:
            _logger.error(f"Fail to compute a score for {url.path()}. Reason is {e}")
            return {cor: (url, None)}
        else:
            if data.ndim == 3:
                if data.shape[0] == 1:
                    data = data.reshape(data.shape[1], data.shape[2])
                elif data.shape[2] == 1:
                    data = data.reshape(data.shape[0], data.shape[1])
                else:
                    raise ValueError(f"Data is expected to be 2D. Not {data.ndim}D")
            elif data.ndim == 2:
                pass
            else:
                raise ValueError("Data is expected to be 2D. Not {data.ndim}D")
            return {cor: (url, data)}

    def load_datasets(self):
        with Pool(self.pool_size) as pool:
            res = pool.map(
                self._load_dataset,
                self._cor_reconstructions.items(),
            )
        datasets_ = {}
        for mydict in res:
            datasets_.update(mydict)
        return datasets_

    def cancel(self):
        self._cancelled = True


class SAAxisProcess(SAAxisTask):
    def __init__(
        self, process_id=None, inputs=None, varinfo=None, node_attrs=None, execinfo=None
    ):
        deprecated_warning(
            name="tomwer.core.process.reconstruction.saaxis.SAAxisProcess",
            type_="class",
            reason="improve readibility",
            since_version="1.2",
            replacement="SAAxisTask",
        )
        super().__init__(process_id, inputs, varinfo, node_attrs, execinfo)



from silx.gui import qt
from silx.gui import icons as silx_icons
from tomwer.gui import icons as tomwer_icons


class LoadConfigurationAction(qt.QAction):
    """Action to trigger load of a stitching configuration"""

    def __init__(self, parent=None, text=None):
        super().__init__(parent)
        if text is not None:
            self.setText(text)
        self.setToolTip("load nabu-stitching configuration")
        load_icon = silx_icons.getQIcon("document-open")
        self.setIcon(load_icon)


class SaveConfigurationAction(qt.QAction):
    """Action to trigger save of a stitching configuration"""

    def __init__(self, parent=None, text=None):
        super().__init__(parent)
        if text is not None:
            self.setText(text)
        self.setToolTip("save nabu-stitching configuration")
        save_icon = silx_icons.getQIcon("document-save")
        self.setIcon(save_icon)


class AddTomoObjectAction(qt.QAction):
    """Action to trigger add a volume to a stitching configuration"""

    def __init__(self, parent=None, text=None):
        super().__init__(parent)
        if text is not None:
            self.setText(text)
        self.setToolTip(
            "Add a NXtomo or a (nabu) reconstructed volume to the stitching"
        )
        add_icon = tomwer_icons.getQIcon("add")
        self.setIcon(add_icon)


class PreviewAction(qt.QAction):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setToolTip("Compute preview with current settings (shortcut: F5)")
        update_preview_icon = tomwer_icons.getQIcon("update_stitching_preview")
        self.setIcon(update_preview_icon)



from __future__ import annotations

import logging

from orangewidget import gui
from orangewidget.settings import Setting
from silx.gui import qt
from ewokscore.missing_data import MISSING_DATA

from ...orange.managedprocess import TomwerWithStackStack

from tomwer.core.scan.nxtomoscan import NXtomoScan
import tomwer.core.process.edit.nxtomoeditor
from tomwer.gui.edit.nxtomoeditor import NXtomoEditorDialog as _NXtomoEditorDialog

from ...orange.managedprocess import _SuperviseMixIn

_logger = logging.getLogger(__name__)


class NXtomoEditorDialog(_NXtomoEditorDialog):

    def __init__(self, parent, *args, **kwargs) -> None:
        assert isinstance(parent, _SuperviseMixIn)
        super().__init__(parent, *args, **kwargs)


class NXtomoEditorOW(
    TomwerWithStackStack,
    ewokstaskclass=tomwer.core.process.edit.nxtomoeditor.NXtomoEditorTask,
):
    """
    Widget to edit manually a NXtomo
    """

    name = "nxtomo-editor"
    id = "orange.widgets.tomwer.edit.NXtomoEditorOW.NXtomoEditorOW"
    description = "Interface to edit manually a NXtomo"
    icon = "icons/nx_tomo_editor.svg"
    priority = 10
    keywords = [
        "hdf5",
        "nexus",
        "tomwer",
        "file",
        "edition",
        "NXTomo",
        "editor",
        "energy",
        "distance",
        "pixel size",
    ]
    _ewoks_inputs_to_hide_from_orange = ("configuration",)

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    sigScanReady = qt.Signal(str)
    """emit when scan ready. Used for test only"""

    settings = Setting(dict())

    def __init__(self, parent=None, *args, **kwargs):
        super().__init__(parent=parent, *args, **kwargs)

        _layout = gui.vBox(self.mainArea, self.name).layout()
        self.widget = NXtomoEditorDialog(parent=self, hide_lockers=False)
        _layout.addWidget(self.widget)
        self.widget.mainWidget.sigEditingFinished.connect(self._updateSettings)

        # load settings
        self.widget.setConfiguration(self.settings)

        # connect signal / slot
        self.widget._buttons.button(qt.QDialogButtonBox.Ok).released.connect(
            self.execute_ewoks_task
        )

    def _updateSettings(self):
        self.settings = self.getConfiguration()

    def handleNewSignals(self) -> None:
        scan = self.get_task_input_value("data", MISSING_DATA)
        if scan not in (MISSING_DATA, None):
            self._setScan(scan=scan)
        # return super().handleNewSignals() do not call to make sure the processing is not triggered

    def setScan(self, scan: NXtomoScan | None):
        self.set_dynamic_input("data", scan)
        self._setScan(scan=scan)

    def _setScan(self, scan: NXtomoScan | None):
        if not isinstance(scan, NXtomoScan):
            raise TypeError(
                f"expect to have an instance of {NXtomoScan}. {type(scan)} provided."
            )

        self.widget.setScan(scan)
        if self.widget.hasLockField():
            super().execute_ewoks_task()
        else:
            self.show()
            self.raise_()

    def task_output_changed(self) -> None:
        self.sigScanReady.emit(str(self.get_task_output_value("data")))
        super().task_output_changed()

    def get_task_inputs(self):
        task_inputs = super().get_task_inputs()
        task_inputs.update(
            {
                "configuration": self.widget.getConfigurationForTask(),
            }
        )
        return task_inputs

    def sizeHint(self):
        return qt.QSize(400, 500)

    # expose API
    def getConfiguration(self) -> dict:
        return self.widget.getConfiguration()

    def setConfiguration(self, config: dict):
        self.widget.setConfiguration(config)



# coding: utf-8

"""Unit test for the scan defined at the hdf5 format"""
from __future__ import annotations

import asyncio
import os

from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.utils.scanutils import MockNXtomo


def test_simple_future_tomo_obj(tmpdir):
    """Simple test of the FutureTomwerScan API"""
    scan = MockNXtomo(
        scan_path=os.path.join(tmpdir, "scan_test"),
        n_proj=10,
        n_ini_proj=10,
        create_ini_dark=False,
        create_ini_flat=False,
        dim=10,
    ).scan
    future = asyncio.Future()
    future.set_result(None)

    future_tomo_obj = FutureTomwerObject(
        tomo_obj=scan,
        futures=[
            future,
        ],
    )

    future_tomo_obj.results()
    assert future_tomo_obj.exceptions() is None
    assert future_tomo_obj.cancelled() is False



from silx.gui import qt

from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.tests.conftest import qtapp  # noqa F401


def test_get_splash_screen(qtapp):  # noqa F401
    assert isinstance(getMainSplashScreen(), qt.QSplashScreen)



# coding: utf-8

from tomwer.tests.conftest import qtapp  # noqa F401
from orangecontrib.tomwer.widgets.control.AdvancementOW import AdvancementOW


def test_totsaso(qtapp):  # noqa F401
    AdvancementOW()
